local MSG = require "api_msg"
local CO = require "api_coservice"
local cjson = require "cjson"
local _dbg = require "coaf.debug"

local dset = {}

local function API_list(red, args)
	local proxy = CO.GetProxy( "/sip", "codec" )
	if not proxy then
		MSG.ERROR( "1200" )
		return true
	else
		local r, rslt = proxy.GetAccountList()
		proxy:Destroy()

		if not r or not rslt then
			MSG.ERROR( "1201" )
			return true
		else
			MSG.OK( "", { data = rslt } )
			return true
		end
	end
end

local function API_add(red, args)
	if not args or not args.name or args.name == "" or not args.type or args.type == "" then
		MSG.ERROR( "1000" )
		return true
	end

	local proxy = CO.GetProxy( "/sip", "codec" )
	if not proxy then
		MSG.ERROR( "1200" )
		return true
	else
		local r, rslt = proxy.AddAccount({ name = args.name, protocol = args.type })
		proxy:Destroy()

		if not r or not rslt then
			MSG.ERROR( "1202" )
			return true
		else
			MSG.OK( "" )
			return true
		end
	end
end

local function API_modify(red, args)
	if not args or not args.id or args.id == "" then
		MSG.ERROR( "1000" )
		return true
	end

	local proxy = CO.GetProxy( "/sip", "codec" )
	if not proxy then
		MSG.ERROR( "1200" )
		return true
	else
		local r, rslt = proxy.ModifyAccount( tonumber(args.id), { 
			name = args.name, 
			protocol = args.type,
			localAccount = args.localAccount,
			localAddress = args.localAddress,
			realm = args.realm,
			serverId = args.serverId,
			registerServer = args.registerServer,
			registerPort = tonumber(args.registerPort),
			username = args.username,
			password = args.password
		})
		proxy:Destroy()

		if not r or not rslt then
			MSG.ERROR( "1203" )
			return true
		else
			MSG.OK( "" )
			return true
		end
	end
end

local function API_remove(red, args)
	if not args or not args.id or args.id == "" then
		MSG.ERROR( "1000" )
		return true
	end

	local proxy = CO.GetProxy( "/sip", "codec" )
	if not proxy then
		MSG.ERROR( "1200" )
		return true
	else
		local r, rslt = proxy.RemoveAccount( tonumber(args.id) )
		proxy:Destroy()

		if not r or not rslt then
			MSG.ERROR( "1204" )
			return true
		else
			MSG.OK( "" )
			return true
		end
	end
end

local function API_status(red, args)
	if not args or not args.id or args.id == "" then
		MSG.ERROR( "1000" )
		return true
	end

	local proxy = CO.GetProxy( "/sip", "codec" )
	if not proxy then
		MSG.ERROR( "1200" )
		return true
	else
		local r, rslt = proxy.GetAccountStatus( tonumber(args.id) )
		proxy:Destroy()

		if not r or not rslt then
			MSG.ERROR( "1205" )
			return true
		else
			local data = {
				{ id="account_status", state=rslt.account_status or "error" },
				{ id="register_status", state=rslt.register_status or "error", extra_info=rslt.register_code },
				{ id="call_status", state=rslt.call_status or "none" },
				{ id="bitrate", state=0 }
			}
			MSG.OK( "", { data = data } )
			return true
		end
	end
end

local function API_server_port(red, args)
	local proxy = CO.GetProxy( "/sip", "codec" )
	if not proxy then
		MSG.ERROR( "1200" )
		return true
	else
		local r, __, rslt = proxy.GetLocalBinds()
		proxy:Destroy()

		if not r or not rslt then
			MSG.ERROR( "1206" )
			return true
		else
			local data = {}
			local k,v
			for k,v in pairs(rslt) do
				table.insert( data, { name=v.bind, value=v.bind } )
			end
			MSG.OK( "", { data = data } )
			return true
		end
	end
end

local function API_account(red, args)
	if not args or not args.id or args.id == "" then
		MSG.ERROR( "1000" )
		return true
	end

	local proxy = CO.GetProxy( "/sip", "codec" )
	if not proxy then
		MSG.ERROR( "1200" )
		return true
	else
		local r, rslt = proxy.GetAccountOptions( tonumber(args.id) )
		proxy:Destroy()

		if not r or not rslt then
			MSG.ERROR( "1207" )
			return true
		else
			MSG.OK( "", { data = rslt } )
			return true
		end
	end
end

local function API_call_cfg(red, args)
	if not args or not args.id or args.id == "" then
		MSG.ERROR( "1000" )
		return true
	end

	local proxy = CO.GetProxy( "/sip", "codec" )
	if not proxy then
		MSG.ERROR( "1200" )
		return true
	else
		local r, rslt = proxy.GetAccountOptions( tonumber(args.id) )
		proxy:Destroy()

		if not r or not rslt then
			MSG.ERROR( "1207" )
			return true
		else
			MSG.OK( "", { data = { 
				allow_in = rslt.call_in_options and rslt.call_in_options.allow and 1 or 0,
				allow_in_num = rslt.call_in_options and rslt.call_in_options.policy or 1,
				auto_aws = rslt.call_in_options and rslt.call_in_options.auto_answer and 1 or 0,
				sip_out_url = rslt.call_out_options and rslt.call_out_options.target or "",
				auto_call = rslt.call_out_options and rslt.call_out_options.auto_call and 1 or 0,
				buffer = rslt.buffer_option or "live:50:50:200:0:0",
				source_enable = rslt.call_out_options and rslt.call_out_options.create_stream and 1 or 0,
				source_id = rslt.source_id or "",
				default_name = rslt.source_id or ""
			}} )
			return true
		end
	end
end

local function API_modify_call_cfg(red, args)
	if not args or not args.id or args.id == "" then
		MSG.ERROR( "1000" )
		return true
	end

	local proxy = CO.GetProxy( "/sip", "codec" )
	if not proxy then
		MSG.ERROR( "1200" )
		return true
	else
		local r, rslt = proxy.ModifyAccount( tonumber(args.id), { 
			buffer_option = args.buffer,
			source_id = args.source_id,
			call_in_options = {
				allow = tonumber(args.allow_in) == 1,
				policy = tonumber(args.allow_in_num),
				auto_answer = tonumber(args.auto_aws) == 1
			},
			call_out_options = {
				target = args.sip_out_url,
				auto_call = tonumber(args.auto_call) == 1,
				create_stream = tonumber(args.source_enable) == 1
			}
		})
		proxy:Destroy()
		if not r or not rslt then
			MSG.ERROR( "1203" )
			return true
		else
			MSG.OK( "" )
			return true
		end
	end
end

local function API_get_binds(red, args)
	local proxy = CO.GetProxy( "/sip", "codec" )
	if not proxy then
		MSG.ERROR( "1200" )
		return true
	else
		local r, rslt = proxy.GetLocalBinds()
		proxy:Destroy()

		if not r or not rslt then
			MSG.ERROR( "1208" )
			return true
		else
			MSG.OK( "", { data = { binds = rslt } } )
			return true
		end
	end
end

local function API_set_binds(red, args)
	if not args or not args.binds or args.binds == "" then
		MSG.ERROR( "1209" )
		return true
	end

	local proxy = CO.GetProxy( "/sip", "codec" )
	if not proxy then
		MSG.ERROR( "1200" )
		return true
	else
		local r = proxy.SetLocalBinds(args.binds)
		proxy:Destroy()

		if not r then
			MSG.ERROR( "1210" )
			return true
		else
			MSG.OK("")
			return true
		end
	end
end

--------------------------------------
local function API_answer(red, args)
	if args and args.id then
		local proxy = CO.GetProxy( "/sip", "codec" )
		if not proxy then
			MSG.ERROR( "1200" )
			return true
		else
			local r,done 
			if args.dir and args.dir == "sip_out" then
				r,done = proxy.MakeCall( args.id )
			else
				r,done = proxy.AnswerCall( args.id )
			end

			proxy:Destroy()
			if not r or not done then
				-- MSG.ERROR( "呼叫/应答失败:" .. (done or "Unknown error") )
				MSG.ERROR( "1211")
			else
				MSG.OK()
			end
			return true
		end
	else
		MSG.ERROR( "1000" )
	end
	return true
end

--------------------------------------

local function API_hangup(red, args)
	if args and args.id then
		local proxy = CO.GetProxy( "/sip", "codec" )
		if not proxy then
			MSG.ERROR( "1200" )
			return true
		else
			local r,done = proxy.HangupCall( args.id )
			proxy:Destroy()
			if not r or not done then
				MSG.ERROR( "1211" )
			else
				MSG.OK()
			end
			return true
		end
	else
		MSG.ERROR( "1000" )
	end
	return true
end


local function API_get_source(red, args)	
	MSG.ERROR( "0000" )
	return true
end


dset.APIS = dset.APIS or {}

dset.APIS["list"] = API_list
dset.APIS["add"] = API_add
dset.APIS["modify"] = API_modify
dset.APIS["remove"] = API_remove
dset.APIS["account"] = API_account
dset.APIS["call_cfg"] = API_call_cfg
dset.APIS["modify_call_cfg"] = API_modify_call_cfg
dset.APIS["get_binds"] = API_get_binds
dset.APIS["set_binds"] = API_set_binds
dset.APIS["server_port"] = API_server_port
dset.APIS["status"] = API_status
dset.APIS["answer"] = API_answer
dset.APIS["hangup"] = API_hangup
dset.APIS["get_source"] = API_get_source

return dset

