local coaf = require "coaf.core"
local _dbg = require "coaf.debug"
local MD = require "audioMixer.mixer_driver"
local TAP = require "audioMixer.tap"

require "kmCore"

local _M = {
	all_mixers = {}
}

----------------------------------

local mixer_meta = {}

function mixer_meta:Init()
	return true
end

--Create mix output:
--Arguments:
--		name : the mixer name. if nil, it use current mixer object name;
--		mixer_type: The mixer type (module name for pulse, or mixer type ID for other mixer, or may be no effect);
--		mixer_args: The mixer arguments (aka format=..., rate=..., channels=..., etc)
--		sink_type: (optional). If the mixer is not driven a physical audio device directly,
--		           (e.g. pulse-null-sink,...). For this case, you can link an actual audio
--		           output sink to this mixer. Normally, this type may be 'ALSA' but you
--		           can specify any supported AudioSink type (e.g. HISIAO, CROSS, ... )
--		sink_args: (optional). The actual audio output sink creation arguments.

function mixer_meta:CreateMixOutput(name, mixer_type, mixer_args, sink_type, sink_args)
	name = name or self.name
	if self.mix_out_taps[name] then
		_dbg.Warn( "MIXER", "CreateMixOutput: The creating mix-output of name:", name, "is already exist! For:", self.name )
		return true
	end

	mixer_type, mixer_args = self.MIXER_DRV.CheckTypeAndArgs(mixer_type, mixer_args)
	if not mixer_type or not mixer_args then
		_dbg.Error( "MIXER", "CreateMixOutput: Invalid mixer type or mixer arguments while creating mixer:", name, "! For:", self.name )
		return false
	end

	if not self.MIXER_DRV.CreateMixer( name, mixer_type, mixer_args ) then
		--_dbg.Error( "MIXER", "CreateMixOutput: Fail to create mixer (driver:", mixerDriver, " name:", name, ",type:", mixer_type, ") for:", self.name )
		_dbg.Error( "MIXER", "CreateMixOutput: Fail to create mixer (name:", name, ",type:", mixer_type, ") for:", self.name )
		return false
	end

	self.mix_out_taps[name] = TAP.Create( ".MXOUT:" .. self.name .. "-" .. name )
	if not self.mix_out_taps[name] then
		self.MIXER_DRV.RemoveMixer( name )
		return false
	end
	self.mix_out_taps[name]:Init()
	self.mix_out_taps[name]:SetOwnerMixer( self, name )

	if sink_type then
		self.mix_out_taps[name]:AddOutputTarget( ".default", sink_type, sink_args )
	end

	if self.started then
		self.mix_out_taps[name]:Enable()
	end

	self.mix_out_args[name] = mixer_args
	return true
end

function mixer_meta:RemoveMixOutput(name)
	if not name or not self.mix_out_taps[name] then
		_dbg.Error("MIXER", "RemoveMixOutput: Invalid mix-output name or it's not created! Name:", name, "for mixer:", self.name )
		return false
	end

	local k,v
	for k,v in pairs(self.taps) do
		v:DisconnectFromMixOutput(name)
	end
	self.mix_out_taps[name]:Destroy()
	self.mix_out_taps[name] = nil
	self.mix_out_args[name] = nil
	self.mix_out_srcs[name] = nil	--Directly remove it, since it should be auto removed.

	self.MIXER_DRV.RemoveMixer(name)
	return true
end

function mixer_meta:RemoveAllMixOutput()
	local allmo = {}
	local k,v
	for k,v in pairs(self.mix_out_taps) do
		table.insert(allmo, k)
	end
	for k,v in pairs(allmo) do
		self:RemoveMixOutput(v)
	end
end

function mixer_meta:GetMixOutputTap(name)
	return self.mix_out_taps[name]
end

function mixer_meta:Destroy()
	self:Stop()
	local all_taps = {}
	local k,v
	for k,v in pairs(self.taps) do
		table.insert( all_taps, k )
	end

	for k = 1, #all_taps do
		self:FreeTap(all_taps[k] )
	end

	self:RemoveAllMixOutput()
end

function mixer_meta:CreateTap(name)
	if not name or name:match("%.MO:") then
		_dbg.Error( "MIXER", "CreateTap: Invalid tap or tap name could not be start with '.MO:'. For:", self.name )
		return nil
	end

	local tap = self.taps[name]
	if tap then
		return tap
	end

	tap = TAP.Create(name)
	if not tap then
		_dbg.Error( "MIXER", "CreateTap: Fail to create tap:", name, "for mixer:", self.name )
		return nil
	end
	tap:Init()
	tap:SetOwnerMixer(self)
	self.taps[name] = tap
	return tap
end

function mixer_meta:GetTap(name)
	if not name then
		_dbg.Error( "MIXER", "GetTap: No valid tap name specified! For:", self.name )
		return nil
	else
		tap = self.taps[name]
		if not tap then
			_dbg.Warn( "MIXER", "GetTap: No tap of name:", name, "exists! For:", self.name )
			return nil
		end
		return tap
	end
end

function mixer_meta:FreeTap(name)
	if not name then
		_dbg.Error( "MIXER", "FreeTap: Invalid tap name. For:", self.name )
		return false
	end
	tap = self.taps[name]
	if tap then
		tap:Destroy()
		self.taps[name] = nil
	end
	return true
end

function mixer_meta:Start()
	if not self.started then
		self.started = true
		local k,v
		for k,v in pairs(self.mix_out_taps) do
			v:Enable()
		end
	end
end

function mixer_meta:Stop()
	if self.started then
		self.started = false
		local k,v
		for k,v in pairs(self.mix_out_taps) do
			v:Disable()
		end
	end
end

--Bellow function is internal using only.
function mixer_meta:_CreateEncoderMixer( tap_name, mixOutName, isMoTap )
	if isMoTap then
		return self.MIXER_DRV.GetChannelIndivMixerSourceTypeAndArgs( mixOutName, true ) --Get table format.
	elseif not mixOutName or not self.mix_out_taps[mixOutName] then
		return nil, "Invalid mix-output name"
	else
		local mixer_name = mixOutName .. "_" .. tap_name
		if self.tap_mixers[mixer_name] then
			return self.MIXER_DRV.GetChannelIndivMixerSourceTypeAndArgs(mixer_name,true)
		end
		if #self.tap_mixers >= self.MIXER_DRV.support_max_tap_mixers then
			_dbg.Error( "MIXER", "Create tap mixer but it is up to max limitation of", self.MIXER_DRV.support_max_tap_mixers, ", mixer:", self.name )
			return nil, "Tap mixers is up to max limitation of " .. self.MIXER_DRV.support_max_tap_mixers
		end

		local done, output_target, output_target_args = self.MIXER_DRV.CreateChannelIndivMixer( mixOutName, tap_name, mixer_name, self.mix_out_args[mixOutName] )
		if not done then
			_dbg.Error( "MIXER", "Fail to create tap pulse mixer for:", self.name )
			return nil, "Fail to create tap pluse mixer"
		end

		table.insert( self.tap_mixers, mixer_name )
		self.tap_mixers[mixer_name] = true

		if output_target then
			local k, tap
			for k,tap in pairs(self.taps) do
				if k ~= tap_name and tap.connectedMixOutputs[mixOutName] then --Don't push self to mixer.
					tap:AddOutputTarget( ".mixer:" .. mixer_name, output_target, output_target_args )
				end
			end
		end
		return self.MIXER_DRV.GetChannelIndivMixerSourceTypeAndArgs(mixer_name,true)
	end
end

--Bellow function is internal using only.
function mixer_meta:_FreeEncoderMixer( tap_name, mixOutName, mixer_name )
	if not mixer_name then return end
	if self.tap_mixers[mixer_name] then
		local k,tap
		for k,tap in pairs(self.taps) do
			if tap.connectedMixOutputs[mixOutName] then
			--NOTE: not very good here, but if that output target is not exist, nothing will happen.
				tap:RemoveOutput( ".mixer:" .. mixer_name )
			end
		end
		self.MIXER_DRV.RemoveChannelIndivMixer( mixOutName, tap_name, mixer_name )
		local i
		for i=1, #self.tap_mixers do
			if self.tap_mixers[i] == mixer_name then
				table.remove( self.tap_mixers,i)
				break
			end
		end
		self.tap_mixers[mixer_name] = nil
	end
end

--Bellow function is internal using only.
function mixer_meta:_CreateDefaultEncodingChannel(mixOutName,codecArgs, tap_name )
	if not mixOutName or not self.mix_out_taps[mixOutName] then
		_dbg.Error( "MIXER", "Invalid mix-output name specified for creating default encoding channel! For:", self.name )
		return nil
	end
	return self.mix_out_taps[mixOutName]:CreateEncodingChannel(codecArgs)
end

--Bellow function is internal using only.
function mixer_meta:_FreeDefaultEncodingChannel(chan,mixOutName)
	if not mixOutName or not self.mix_out_taps[mixOutName] then
		return false
	end
	return self.mix_out_taps[mixOutName]:FreeEncodingChannel(chan)
end

--Bellow function is internal using only.
function mixer_meta:_RequireMixOutputTapSource(mixOutName)
	if not mixOutName or not self.mix_out_taps[mixOutName] then
		_dbg.Error( "MIXER", "Invalid mix-output name while creating tap source for:", self.name )
		return false
	end


	if not self.mix_out_srcs[mixOutName] then
		local src_type, src_args = self.MIXER_DRV.GetMixerSourceTypeAndArgs( mixOutName )
		if not src_type then
			return true
		end

		self.mix_out_srcs[mixOutName] = KM_STREAMING.AudioEngine.CreateAudioSource(src_type, src_args)
		if not self.mix_out_srcs[mixOutName] then
			_dbg.Error( "MIXER", "Fail to create mix-output tap source of:", mixOutName, "for mixer:", self.name )
			return false
		end
		self.mix_out_taps[mixOutName]:SelectInputSourceObject( self.mix_out_srcs[mixOutName] )
		return true
	end
	return true
end

--Bellow function is internal using only.
function mixer_meta:_ReleaseMixOutputTapSource(mixOutName)
	if mixOutName and self.mix_out_srcs[mixOutName] then
		self.mix_out_taps[mixOutName]:RemoveInput()
		KM_STREAMING.AudioEngine.FreeAudioSource( self.mix_out_srcs[mixOutName] )
		self.mix_out_srcs[mixOutName] = nil
	end
end

mixer_meta.__index = mixer_meta

----------------------------------

function _M.Create(name, mixerDriver)
	if not name then
		_dbg.Error( "MIXER", "Create: You must specify a valid name for creating mixer!" )
		return nil
	end

	mixerDriver = mixerDriver or MD.default_mixer_driver
	local MIXER_DRV = MD.supported_mixer_drivers[ mixerDriver ] 
	if not MIXER_DRV then
		_dbg.Error( "MIXER", "Create: Invalid mixer driver:", mixerDriver, "specified!" )
		return nil
	end
	
	if _M.all_mixers[name] then
		_dbg.Warn( "MIXER", "Create: mixer '" .. name .. "' is already exist!" )
		return _M.all_mixers[name]
	end

	local obj = {
		started = false,
		name = name,
		MIXER_DRV = MIXER_DRV,
		mix_out_taps = {},
		mix_out_args = {},
		mix_out_srcs = {},
		taps = {},
		tap_mixers = {},
		encodingChannels = {}
	}

	setmetatable(obj, mixer_meta)
	if not obj:Init() then
		_dbg.Error( "MIXER", "Create: Fail to initialize mixer:", name )
		obj = nil
		return nil
	end

	_M.all_mixers[name] = obj
	return obj
end

return _M

