local coaf = require "coaf.core"
local _dbg = require "coaf.debug"

local TOOL = "/usr/bin/pactl"

local DEFAULT_DELAY = 30

local _M = {
	support_max_tap_mixers = 16
}

function _M.CheckTypeAndArgs(module_type, args)
	module_type = module_type or "module-null-sink"
	
	if module_type == "default" or module_type == "DEFAULT" or module_type == "null" or module_type == "NULL" or module_type == "" then
		module_type = "module-null-sink"
	elseif module_type == "alsa" or module_type == "ALSA" then
		module_type = "module-alsa-sink"
	end

	if not module_type:match( "module%-.*%-sink" ) then
		_dbg.Error( "PULSE", "Checking: invalid module type of:", module_type, "!" )
		return nil, nil
	end

	args = args or 'format="s16le" rate=48000 channels=2'
	if module_type == "module-alsa-sink" then
		if not args:match("device=") then
			args = args .. ' device="sysdefault"'
		end
	end
	return module_type, args
end

function _M.CheckMixerConnection( tap_name, mixer_name )
	return true
end

function _M.ReleaseMixerConnection( tap_name, mixer_name )
end

function _M.CreateMixer(name, module_type, args)
	if not name then
		_dbg.Error( "PULSE", "Invalid mixer name while creating mixer!" )
		return nil
	end

	local id = _M.MixerExists(name)
	if id then
		_dbg.Warn( "PULSE", "The mixer '", name, "' is already exist, id=", id )
		return id
	end

	--module_type, args = _M.CheckTypeAndArgs(module_type,args)

	local r,stat,code = os.execute( TOOL .. ' load-module ' .. module_type .. ' sink_name="' .. name .. '" ' .. args )
	if not r or code ~= 0 then
		_dbg.Error( "PULSE", "Fail to create a pulse mixer for name:", name, "type:", module_type, "args:", args )
		return nil
	else
		return _M.MixerExists(name)
	end
end

function _M.CreateChannelIndivMixer(base_mixer_name, channel_name, suggest_mixer_name, args)
	if not _M.CreateMixer( suggest_mixer_name, "module-null-sink", args ) then
		return false, nil, nil
	else
		return true, "ALSA", { device = "mixer:" .. suggest_mixer_name, buffer = DEFAULT_DELAY }
	end
end

function _M.GetChannelIndivMixerSourceTypeAndArgs(mixer_name, tableFmt)
	if tableFmt then
		return "ALSA", { device = "mixer:" .. mixer_name .. ".monitor" }
	else
		return { type="ALSA", device = "mixer:" .. mixer_name .. ".monitor", mixer_name = mixer_name }
	end
end

function _M.RemoveChannelIndivMixer( base_mixer_name, channel_name, mixer_name )
	return _M.RemoveMixer(mixer_name)
end

function _M.GetMixerSourceTypeAndArgs(mixer_name)
	return "ALSA", { device = "mixer:" .. mixer_name .. ".monitor" }
end

function _M.GetMixerSinkTypeAndArgs(mixer_name)
	return "ALSA", { device = "mixer:" .. mixer_name, buffer = DEFAULT_DELAY }
end

function _M.RemoveMixer(name)
	if not name then
		_dbg.Error( "PULSE", "Invalid mixer name while removing mixer!" )
		return nil
	end

	local id,mtype = _M.MixerExists(name)
	if not id then
		_dbg.Warn( "PULSE", "The specified mixer name:", name, "is already removed." )
		return true
	end
	
	--Don't unload alsa sink driver. But need improve this action later.
	if mtype == "module-alsa-sink" then
		return true
	end

	local r,stat,code = os.execute( TOOL .. ' unload-module ' .. id )
	if not r or code ~= 0 then
		_dbg.Error( "PULSE", "Fail to remove a pulse mixer for name:", name )
		return false
	else
		return true
	end
end

function _M.GetMixers()
	local f = io.popen( TOOL .. " list short modules", "r" )
	if not f then
		return nil
	else
		local rslt = {}
		local ln = f:read( "*l" )
		while ln do
			local id, mo_type, name = ln:match('^(%d+)%s+(module%-.*)%s+.*sink_name%s*=%s*"(.-)"')
			if not id then --Try another.
				id, mo_type, name = ln:match('^(%d+)%s+(module%-.*)%s+.*sink_name%s*=%s*([^%s]+)')
			end
			if id and name then
				table.insert( rslt, { id = tonumber(id), mo_type = mo_type, name = name } )
			end
			ln = f:read("*l")
		end
		f:close()
		return rslt
	end
end

function _M.MixerExists(name)
	local r = _M.GetMixers()
	if r then
		local i
		for i=1, #r do
			if r[i].name == name then
				return r[i].id, r[i].mo_type
			end
		end
		return nil
	else
		return nil
	end
end

return _M


