local sys_config = require "sys_config"
local sys_video_config = require "sys_video_config"
local sys_audio_config = require "sys_audio_config"
local coaf = require "coaf.core"
local _dbg = require "coaf.debug"
local output = require "soc.output"

local layout = require "mod_Layout"
local streams = require "mod_Streams"

local cfg_output = require "configs.output_config"
local vo_drivers = require "video.vo_drivers"
local CO = require "mod_coservice"
local pt = require "pl.pretty"
local mo_OSD_adap = require 'mod_OSD_adapter'
require "himpp"
local VPSS = HiMPP.VPSS

local _done, features = pcall(require, "features")
if not _done then
	features = {}
elseif type(features) ~= "table" then
	features = _G
end

local function getimgSize(imgpath)
    local proxy = CO.GetProxy('/pics', 'picManager')
    if not proxy then
        _dbg.Error('IMGO', 'Connect picManager service error')
        return 0, 0
    else
        local r, size_w, size_h = proxy.GetPicRate(imgpath)
        proxy:Destroy()
        if not r or not size_w or not size_h then
            _dbg.Error('IMGO', '获取图片尺寸失败！')
            return 0, 0
        else
            return size_w, size_h
        end
    end
    return 0, 0
end

local function convertImg(src, dst)
    local proxy = CO.GetProxy('/pics', 'picManager')
    if not proxy then
        _dbg.Error('IMGO', 'Connect picManager service error')
        return false
    else
        local r, Data = proxy.ConvertPic(src, dst)
        proxy:Destroy()
        if not r or not Data then
            _dbg.Error('IMGO', '转换图片失败！')
            return false
        else
            return Data
        end
    end
    return false
end

local function GetHomePicInfos()
	local proxy = CO.GetProxy('/relation', 'picManager')
    if not proxy then
        _dbg.Error('IMGO', '无法连接图片服务！')
        return nil
	else
        local r, Data = proxy.GetHomePicInfos()
        proxy:Destroy()
        if not r or not Data then
            _dbg.Error('IMGO', '获取图片信息失败！')
            return nil
		else
            return Data
        end
    end
    return nil
end

local ALL_POSSIBLE_INTFS = {
	"HDMI",
	"BT1120",
	"SDI",
	"VGA",
	"CVBS",
	"NET"
}

local _M = {
	audio_output_mapping = sys_video_config.outputs.audio_mapping,
	enabled_interfaces = sys_video_config.outputs.enabled_interfaces,
	--[[ Like this:
	{
		["HDMI"] = "HDMI",
		["SDI"] = "SDI",
		["BT1120"] = "SDI",
		["LINEOUT"] = "LINEOUT",
	}]]
	all_outputs = {},
	BG_VPSS_ID_BASE = 18
}

function _M.Init()
	_M.audio_out_config = require "configs.audio_out_config".Load()
end

local function OSDChangeResolution(output, width, height) 
    local proxy = CO.GetProxy( "/osd", "OSD" )
    if not proxy then
        _dbg.Error("OUTPUT", "无法连接OSD服务！" )
        return false
    else
		local r,Data = proxy.ChangeResolution(output,width,height)
		_dbg.Info("OUTPUT", "OSD ChangeResolution",r,Data )
        proxy:Destroy()
        if not r or not Data then
            _dbg.Error("OUTPUT", "修改OSD分辨率失败" )
            return false
        else
            return Data
        end
    end	
    return false   
end

function _M.OSDSetBackGround(enable, output, img_id)
	-- print("OSDSetBackGround~~~~~~~~~~~~~~~~~~~~",output,img_id)
	img_id = tonumber(img_id)

	if not output or not img_id then
		_dbg.Error("OUTPUT", "SetBackGround, but no img_id/output date" )
		return false
	end
	
	local outputid, output_device
	if output == 'hd0' then
		outputid = 1
		output_device = 'hd:0'
	else
		outputid = 2
		output_device = 'hd:1'
	end

	local out = _M.all_outputs[output_device]
	if not out then
		_dbg.Error("OUTPUT", "output object get error!" )
		return false
	end

	local proxy_img = CO.GetProxy( "/pics", "picManager", "local", 10)
	if proxy_img then
		local r, done, err = proxy_img.ModifyPic({
			id = img_id,
			output = outputid,
			type = "bg",
			enable = enable,
		}, true)
		proxy_img:Destroy()
		if r and done then
			return true
		else
			_dbg.Error("OUTPUT", "ModifyPic error!", err )
			return false
		end
	else
		return false
	end
end

-----------------------------------------------------
-- NOTE: Don't make a fuss! This module is changed from
-- an old version that only support ONE output screen.
-- but now it can support multiple. For compatible with
-- old codes, I make the functions same as old version.
-- So, I must make a 'wrapper' by Lua closure.
-----------------------------------------------------

function _M.CreateOutput(output_name, output_device, output_id)
--==================================================
--{ begin of the wrapper:
--==================================================

output_device = output_device or "hd:0"

if _M.all_outputs[output_device] then
	_dbg.Warn( "OUTPUT", "CreateOutput: The specified output device:", output_device, "is already created! return the exist object!" )
	return _M.all_outputs[output_device]
end

local _OUT = {
	dispatcher_config = _M.audio_out_config[output_name] or {},
	name = output_name,
	output_id = output_id,
	device = output_device,
	output_config = nil,
	layout_map = {},
	deprecated_streams = {},
	layout_created = false,
	enabled_interfaces = {},
	virtual_mixer = nil,
	osdinfo = nil,
}

function _OUT.Init(disabledBindIntfs)
	local intfs = {}
	local i
	for i=1, #ALL_POSSIBLE_INTFS do
		if _M.enabled_interfaces[ ALL_POSSIBLE_INTFS[i] ] then
			if disabledBindIntfs then
				if not disabledBindIntfs[ ALL_POSSIBLE_INTFS[i] ] then
					local j
					local disabled
					for j=1, #disabledBindIntfs do
						if disabledBindIntfs[j] == ALL_POSSIBLE_INTFS[i] then
							disabled = true
							break
						end
					end
					if not disabled then
						table.insert( intfs, ALL_POSSIBLE_INTFS[i] )
						intfs[ ALL_POSSIBLE_INTFS[i] ] = _M.enabled_interfaces[ ALL_POSSIBLE_INTFS[i] ]
					end
				end
			else
				table.insert( intfs, ALL_POSSIBLE_INTFS[i] )
				intfs[ ALL_POSSIBLE_INTFS[i] ] = _M.enabled_interfaces[ ALL_POSSIBLE_INTFS[i] ]
			end
		end
	end

	_OUT.enabled_interfaces = intfs

	_OUT.output_config = cfg_output.Load(_OUT.name)
	_OUT.disabled_bind_interfaces = disabledBindIntfs

	if not _OUT.output_config then
		_dbg.Error( "OUTPUT", "Fail to load output configure!" )
		_OUT.output_config = {
			layout = "1",
			resolution = _OUT.device:match("hd:") and "1080P30" or "PAL",
			cvbs_resolution = "PAL",
			colorspace = "AUTO",
			hdmi_mode = "HDMI",
			channels = {},
			analog_audio = {},
			embedded_audio = {},
			__TRIGGER = function()
			end
		}
	elseif _OUT.device:match("sd:") then
		if _OUT.output_config.resolution ~= "PAL" and
		   _OUT.output_config.resolution ~= "NTSC" and
		   _OUT.output_config.resolution ~= "480I" and
		   _OUT.output_config.resolution ~= "576I" and
		   _OUT.output_config.resolution ~= "480I60" and
		   _OUT.output_config.resolution ~= "480I50" then
			_OUT.output_config.resolution = "PAL"
			_OUT.output_config:TAB_RESET_CHANGE()
		end
	end

	--For initialization.
	if not _OUT.output_config.bind_interfaces then
		if _OUT.device:match("sd:") then
			_OUT.output_config.bind_interfaces = sys_video_config.outputs.sd0_default_bind.video
		elseif _OUT.device == "hd:0" then
			_OUT.output_config.bind_interfaces = sys_video_config.outputs.hd0_default_bind.video
		else
			_OUT.output_config.bind_interfaces = sys_video_config.outputs.hd1_default_bind.video
		end
		_OUT.output_config:__TRIGGER(true)
	end

	if not _OUT.output_config.bind_audio_outputs then
		if _OUT.device:match("sd:") then
			_OUT.output_config.bind_audio_outputs = sys_video_config.outputs.sd0_default_bind.audio
		elseif _OUT.device == "hd:0" then
			_OUT.output_config.bind_audio_outputs = sys_video_config.outputs.hd0_default_bind.audio
		else
			_OUT.output_config.bind_audio_outputs = sys_video_config.outputs.hd1_default_bind.audio
		end
		_OUT.output_config:__TRIGGER(true)
	end

	-- Set virtual audio mixer
	if _OUT.device:match("sd:") then
		_OUT.virtual_mixer = sys_video_config.outputs.sd0_default_bind.virtual_mixer
	elseif _OUT.device == "hd:0" then
		_OUT.virtual_mixer = sys_video_config.outputs.hd0_default_bind.virtual_mixer
	else
		_OUT.virtual_mixer = sys_video_config.outputs.hd1_default_bind.virtual_mixer
	end

	_OUT.screen = output.CreateOutput( _OUT.device, _OUT.output_config.resolution, _OUT.output_config.bind_interfaces )

	if not _OUT.screen then
		_dbg.Error( "OUTPUT", "Fail to create output screen object!" )
	else
		_OUT.screen:Init()
		if not features.DISABLE_BACKGROUND then
			_OUT.screen:CreateChannel( 0 )	--bg vochannel
		end
		_OUT.SetOutputFormat()
		_OUT.SetHDMIOptions()
		_OUT.SetOutputLayout() --Using current.
	end

	--Init audio gains.
	_OUT.SetAudioGain( "analog" )
	_OUT.SetAudioGain( "embedded" )
end

function _OUT.setBackGround(url)
	if not url then
		url = ""
	end
	if url == "" then
		_OUT.DisableGroundOSD()
	else
		local w,h = getimgSize(url)
		if not w or not h then
			_dbg.Error('IMGO', 'Get Image width or height error!!!')
			return false
		end

		if convertImg('/data/storage/img/' .. url, '/tmp/fb' .. _OUT.name .. 'bg.rgba') then
			_OUT.bg = url
			_OUT.SetBackGroundOSD(w, h, '/tmp/fb' .. _OUT.name .. 'bg.rgba')
		else
			_dbg.Error('IMGO', 'convert BackGround error')
			return false
		end
	end
	return true
end

function _OUT.UpdatePicInfo()
	local picinfo = GetHomePicInfos()
	if not picinfo then
		_dbg.Error('IMGO', 'Fail to Get HomePic Info!')
		return
	end
	-- coaf.ScheduleDelayedTask(function()
	local i,pic
	for i, pic in ipairs(picinfo) do
		if  pic.output == _OUT.output_id and pic.type == "bg" then
			if (pic.enable == "1" or pic.enable == 1 ) and  pic.url ~= "" then
				if pic.w ~= 0 and pic.h ~=0 then
					-- if _OUT.bg and _OUT.bg ~= pic.url then
						if convertImg('/data/storage/img/' .. pic.url, '/tmp/fb' .. _OUT.name .. 'bg.rgba') then
							-- _OUT.bg = pic.url
							_OUT.SetBackGroundOSD(pic.w, pic.h, '/tmp/fb' .. _OUT.name .. 'bg.rgba')
						else
							_dbg.Error('IMGO', 'convert BackGround error')
						end
					-- else
					-- 	_dbg.Info('IMGO', '_OUT.bg is the same!')
					-- end
				end
			else
				_OUT.DisableGroundOSD()
			end
		end
	end
	-- end, 0)
end

function _OUT.InitBackGround()
	if features.DISABLE_BACKGROUND then
		return true
	end
	local r, errmsg
	local bgvpssid
	if _OUT.device == "hd:0" then
		bgvpssid = 1
	elseif _OUT.device == "hd:1" then
		bgvpssid = 2
	end

	if _OUT.bg_vpss_grp then
		_dbg.Info("OUTPUT","InitBackGround: Destroy exist bg_vpss_grp first")
		_OUT.bg_vpss_grp:Stop()
		_OUT.bg_vpss_grp:Destroy()		
		_OUT.bg_vpss_grp = nil
		collectgarbage()
	end

	VPSS.ForceFreeGroup( _M.BG_VPSS_ID_BASE + bgvpssid )
	_OUT.bg_vpss_grp = VPSS.GetGroup( _M.BG_VPSS_ID_BASE + bgvpssid )
	if not _OUT.bg_vpss_grp then
		_dbg.Error( "OUTPUT", "Fail to create VPSS group, for bg_vpss_grp channel ID=" .. bgvpssid )
		return false
	end

	r, errmsg = _OUT.bg_vpss_grp:Create({
		u32MaxW = 3840,
		u32MaxH = 2176,
		enPixFmt = "YUV_SEMIPLANAR_420",
		--bNrEn = true,
		enDieMode = "NODIE"
	})
	if not r then
		_dbg.Error( "OUTPUT", "Fail to initialize VPSS group, for bg_vpss_grp channel ID=" .. bgvpssid .. ", error:", errmsg )
		_OUT.bg_vpss_grp = nil
		return false
	end

	r = _OUT.bg_vpss_grp:Start()
	if r ~= 0 then
		_dbg.Error( "OUTPUT", "Fail to start VPSS group for bg_vpss_grp(ID=" .. bgvpssid .. "), error:", string.format("%08X",r) )
		return false
	end

	_OUT.bg_vpsschn = _OUT.bg_vpss_grp:GetChannel(0)
	if not _OUT.bg_vpsschn then
		_dbg.Error( "OUTPUT", "Fail to get bg_vpsschn, so bg OSD may be wrong!" )
		return false
	end	

	-- _OUT.screen:BindChannelSource( 0, _OUT.bg_vpsschn, true )
	return true
end

local function file_exists(name)
	local f=io.open(name,"r")
	if f~=nil then io.close(f) 
		return true 
	else 
		return false 
	end
end

function _OUT.DisableGroundOSD()
	if features.DISABLE_BACKGROUND then
		return true
	end
	if _OUT.osdinfo then
		_OUT.osdinfo.bgfilepath = nil
	else
		_OUT.osdinfo = {
			width =  _OUT.screen.syncDetail.width,
			height =  _OUT.screen.syncDetail.height,
			color = 0x000000,
			bgfilepath = nil
		}
	end

	if _OUT.bgosd then
		_OUT.bgosd:EnableArea(0, false)
		_OUT.bgosd:Update()
		-- _OUT.bgosd:Destroy()
		-- _OUT.bgosd = nil
		-- collectgarbage()
	end

	return _OUT.SendUserFrame(_OUT.osdinfo.width, _OUT.osdinfo.height, _OUT.osdinfo.bgcolor )
end

function _OUT.SendUserFrame( width, height, bgcolor )
	if _OUT.bg_vpsschn then
		_OUT.screen:BindChannelSource( 0, _OUT.bg_vpsschn, true, { w=width, h=height } )	
		local r = _OUT.bg_vpss_grp:SendBgColor({
			width = width,
			height = height,
			color = bgcolor,
			s32MilliSec = 0
		})
		if r ~= 0 then
			_dbg.Error( "OUTPUT", "Fail to SendBlackFrame (" .. _OUT.name .. "), error:", string.format("%08X",r) )
			return false
		end
	end
	return true
end

function _OUT.SetBackGroundOSD( width, height, bgfilepath, bgcolor )
	_dbg.Info( "OUTPUT", "_OUT.SetBackGroundOSD start", width, height, bgfilepath, bgcolor)
	if features.DISABLE_BACKGROUND then
		return true
	end
	if not width then  
		width = _OUT.osdinfo and _OUT.osdinfo.width or _OUT.screen.syncDetail.width
	end
	if not height then
		height = _OUT.osdinfo and _OUT.osdinfo.height or _OUT.screen.syncDetail.height
	end

	if not bgcolor then
		bgcolor = _OUT.osdinfo and _OUT.osdinfo.bgcolor or 0x000000
	end

	-- _dbg.Info( "OUTPUT", "SetBackGroundOSD SendBgColor", width, height, bgcolor)
	-- _OUT.SendUserFrame( width, height, bgcolor )

	if not bgfilepath and _OUT.osdinfo and _OUT.osdinfo.bgfilepath then
		bgfilepath = _OUT.osdinfo.bgfilepath
	end

	if not bgfilepath then
		_dbg.Info( "OUTPUT", "SetBackGroundOSD no bgfile ... ", bgfilepath )
		
		_OUT.osdinfo = {
			width = width,
			height = height,
			bgcolor = bgcolor,
			bgfilepath = nil,
		}
		return true
	end 
	
	if not file_exists(bgfilepath) then
		_dbg.Error( "OUTPUT", "SetBackGroundOSD no bgfile miss!... ", bgfilepath )
		return false
	end
	
	if _OUT.bgosd then
		_OUT.bgosd:Destroy()
		_OUT.bgosd = nil
		collectgarbage()
	end

	-- if not enable then
	-- 	_dbg.Error( "OUTPUT", "SetBackGroundOSD disable background" )
	-- 	return true
	-- end

	-- _dbg.Info( "OUTPUT","reinit osd~~~~~", width, height, bgfilepath)
	_OUT.bgosd = mo_OSD_adap.createNew(
		'hi_region',
		{   
			overlay_type = "OVERLAYEX",
			channel = _OUT.bg_vpsschn,
			-- channel = _OUT.screen:GetVoChannel(0).channel,
			base_id = 1,
			width = width,
			height = height,
		}
	)

	if not _OUT.bgosd then
		_dbg.Error('OUTPUT', 'Fail to create backgroud osd !')
		return false
	else
		_dbg.Info('OUTPUT', 'Created backgroud osd !')
		-- osd:SetAreaPositionType(0, "left-top")
		_OUT.bgosd:ChangeAreaType(0, 'image')
		_OUT.bgosd:EnableArea(0, true)
		_OUT.bgosd:ChangeAreaSize(0, width, height)
		_OUT.bgosd:SetAreaPosition(0, 0, 0)
		_OUT.bgosd:SetBackGroundFlag(0, true)
		_OUT.bgosd:LoadFile(0, {filename = bgfilepath, w = width, h = height})
		_OUT.bgosd:Update()
	end

	_OUT.osdinfo = {
		width = width,
		height = height,
		bgcolor = bgcolor,
		bgfilepath = bgfilepath,
	}

	_OUT.SendUserFrame( width, height, bgcolor )

	return true
end

function _OUT.CheckLocalAudioEnable(port)
	local i,v
	if _OUT.virtual_mixer then
		if type(_OUT.virtual_mixer) == "table" then
			for i,v in ipairs(_OUT.virtual_mixer) do
				if v == port then
					return true
				end
			end
		elseif _OUT.virtual_mixer == port then
			return true
		end
	end

	if sys_video_config.outputs.common_audio_bind then
		for i,v in ipairs(sys_video_config.outputs.common_audio_bind) do
			if v == port then
				local atype = sys_audio_config.outputs[v] and sys_audio_config.outputs[v].audio_type
				if atype == "analog" then
					if _OUT.output_config.analog_audio.local_enabled == 1 then
						return true
					else
						return false
					end
				else
					if _OUT.output_config.embedded_audio.local_enabled == 1 then
						return true
					else
						return false
					end
				end
			end
		end
	end

	for i,v in ipairs(_OUT.output_config.bind_audio_outputs) do
		if v == port then
			local atype = sys_audio_config.outputs[v] and sys_audio_config.outputs[v].audio_type
			if atype == "analog" then
				if _OUT.output_config.analog_audio.local_enabled == 1 then
					return true
				else
					return false
				end
			else
				if _OUT.output_config.embedded_audio.local_enabled == 1 then
					return true
				else
					return false
				end
			end
		end
	end

	return false
end

function _OUT.Start()
	if _OUT.screen then
		_OUT.screen:Start()
	end
	_OUT.UpdateOutputEncMode()
end

function _OUT.Stop()
	if _OUT.screen then
		_OUT.screen:Stop()
	end
end

--NOTE: This is a bug patch for Hi3535 (not sure it's exist in other chips). While you
--      first bind a BT1120 interface on the output, it will not output video, util you
--      re-bind again.
function _OUT.PatchBT1120Bug()
	if _OUT.output_config and _OUT.output_config.bind_interfaces then
		local fnd
		local i
		for i=1, #_OUT.output_config.bind_interfaces do
			if _OUT.output_config.bind_interfaces[i] == "BT1120" or
			   _OUT.output_config.bind_interfaces[i] == "SDI" then
				fnd = _OUT.output_config.bind_interfaces[i]
				break
			end
		end
		if not fnd then
			return
		end

		_OUT.UnbindVideoInterface(fnd, {no_osd_update=true, no_encoder_update=true} )
		_OUT.BindVideoInterface(fnd)
	end
end

function _OUT.GetLiveStatusList()
	
end

function _OUT.GetStatus(no_layout)
	-- local layout_number = 1 or layout.
	-- _OUT.output_config.layout
	local layoutDetail = {}

	if ( _OUT.output_config ) then
		layoutDetail = layout.GetLayoutDetail(_OUT.output_config.layout) or {}
	end

	local ret = {
		started = _OUT.screen and _OUT.screen.started or false,
		resolution = _OUT.output_config.resolution,
		resolution_name = _OUT.output_config.resolution,
		layout = _OUT.output_config.layout,
		layout_number =  #layoutDetail,
		bind_interfaces = _OUT.output_config.bind_interfaces,
		hdmi_mode = _OUT.output_config.hdmi_mode,
		colorspace = _OUT.output_config.colorspace,
		sampling = _OUT.output_config.hdmi_audio_sampling,
		border = _OUT.output_config.border
	}

	if not _OUT.cached_support_resolutions then
		_OUT.GetSupportResolutions()
	end

	local act_res = _OUT.screen and _OUT.screen.resolution
	local cfg_res = _OUT.output_config.resolution
	local set_res = cfg_res
	if vo_drivers.specialFormatMap and vo_drivers.specialFormatMap[cfg_res] then
		set_res = vo_drivers.specialFormatMap[cfg_res]
	end

	if act_res and set_res ~= act_res then
		cfg_res = act_res
	end

	local i,v
	for i,v in pairs(_OUT.cached_support_resolutions) do
		if v.value == cfg_res then
			ret.resolution_name = v.name
			break
		end
	end

	if no_layout then
		return ret
	end

	ret.channels = {}

	local i, chn
	local item
	for i,chn in pairs(_OUT.layout_map) do
		local streamObj
		if _OUT.output_config.channels[i] and _OUT.output_config.channels[i].stream then
			streamObj = streams.GetStreamObject( _OUT.output_config.channels[i].stream )
		end

		item = {
			id = i,
			stream = _OUT.output_config.channels[i] and _OUT.output_config.channels[i].stream or "",
			bind = streamObj and true or false
		}

		table.insert( ret.channels, item )
	end

	table.sort( ret.channels, function(a,b)
		return a.id < b.id
	end)

	return ret
end

-- function _OUT.SetOutputFormat(format, needsendbg)
function _OUT.SetOutputFormat(format)
	-- if needsendbg == nil then
	-- 	needsendbg = true
	-- end
	if format then
		_OUT.output_config.resolution = format
		_OUT.output_config:__TRIGGER(true)
	end

	if _OUT.screen then
		_OUT.screen:ChangeResolution(_OUT.output_config.resolution)
		-- 将修改的新分辨率和layout信息传给OSDService
		---- local item = vo_drivers.syncDetailMap[_OUT.screen.resolution]
		--_dbg.Info( "OUTPUT", _OUT.name, "screen w h =",_OUT.screen.syncDetail.width,_OUT.screen.syncDetail.height)
		--if(_OUT.output_config and #_OUT.output_config.bind_interfaces > 0) then
		--	OSDChangeResolution(_OUT.name,_OUT.screen.syncDetail.width,_OUT.screen.syncDetail.height)	
		--end
		_OUT.InitBackGround()
		_OUT.UpdatePicInfo()
		--[[
		coaf.ScheduleDelayedTask(function()
			_OUT.UpdatePicInfo()
		end, 0)
		]]

		_dbg.Info( "OUTPUT", "Output video resolution is changed, update overlay and output encoding...")
		--if #_OUT.output_config.bind_interfaces > 0 then
		OSDChangeResolution(_OUT.name,_OUT.screen.syncDetail.width,_OUT.screen.syncDetail.height)
		--end

		local mo_encoder = require "mod_Encoder"
		mo_encoder.UpdateEncoders()

		----NOTE解决D300,在设置视频格式参数后出现的bug：Select Hi-encoders fd timeout!
		--NOTE D300等系列产品
		if features.FEATURE_IPStream then
			_OUT.Update_D300_Bind_OutputEncMode(false)
		end
	end
end

function _OUT.GetResolution()
	-- 未绑定输出
	if(_OUT.output_config and #_OUT.output_config.bind_interfaces == 0) then
		return nil,nil
	end

	if _OUT.screen and _OUT.screen.syncDetail then 
		return _OUT.screen.syncDetail.width, _OUT.screen.syncDetail.height 
	end
	return nil,nil 
end

function _OUT.GetSupportResolutions()
	if _OUT.cached_support_resolutions then
		return _OUT.cached_support_resolutions
	end

	local Enable4k = true	--EDUO disable 4K
	if features.DEVICE_TYPE == "E_DOCTOR" or features.DEVICE_TYPE == "E_DOCTOR_S" then
		Enable4k = false
	end

	local results
	if _OUT.device == "hd:0" and vo_drivers.support_4k and Enable4k then
		results = {
			{
				name = "(4K)3840x2160@60Hz",
				value = "3840x2160@60"
			},
			{
				name = "(4K)3840x2160@50Hz",
				value = "3840x2160@50"
			},
			{
				name = "(4K)3840x2160@30Hz",
				value = "3840x2160@30"
			},
			{
				name = "(4K)3840x2160@25Hz",
				value = "3840x2160@25"
			},
		} 
	else
		results = {}
	end

	if not _OUT.device:match("sd:") then
		table.insert( results, {
			name = "1920x1080P 60Hz",
			value = "1080P60"
		})
	
		if vo_drivers.support_t1001 then
			table.insert( results, {
				name = "1920x1080P 59.94Hz (SDI only)",
				value = "1080P59"
			})
		end
	
		table.insert( results, {
			name = "1920x1080P 50Hz",
			value = "1080P50"
		})
	
		table.insert( results, {
			name = "1920x1080P 30Hz",
			value = "1080P30"
		})
	
		if vo_drivers.support_t1001 then
			table.insert( results, {
				name = "1920x1080P 29.97Hz (SDI only)",
				value = "1080P29"
			})
		end	

		table.insert( results, {
			name = "1920x1080P 25Hz",
			value = "1080P25"
		})
	
		table.insert( results, {
			name = "1920x1080P 24Hz",
			value = "1080P24"
		})
	
		if vo_drivers.support_t1001 then
			table.insert( results, {
				name = "1920x1080P 23.98Hz (SDI only)",
				value = "1080P23"
			})
		end
	
		table.insert( results, {
			name = "1920x1080i 60Hz",
			value = "1080I60"
		})
	
		if vo_drivers.support_t1001 then
			table.insert( results, {
				name = "1920x1080i 59.94Hz (SDI only)",
				value = "1080I59"
			})
		end
	
		table.insert( results, {
			name = "1920x1080i 50Hz",
			value = "1080I50"
		})
	
		table.insert( results, {
			name = "1280x720P 60Hz",
			value = "720P60"
		})
	
		if vo_drivers.support_t1001 then
			table.insert( results, {
				name = "1280x720P 59.94Hz (SDI only)",
				value = "720P59"
			})
		end
	
		table.insert( results, {
			name = "1280x720P 50Hz",
			value = "720P50"
		})
	end

	if features.DEVICE_TYPE ~= "E_DOCTOR" and features.DEVICE_TYPE ~= "E_DOCTOR_S" then
		table.insert( results, {
			name = "576i/625i 50Hz",
			value = "PAL"
		})
		

		table.insert( results, {
			name = "480i/525i 60Hz",
			value = "NTSC"
		})
	end
	_OUT.cached_support_resolutions = results
	return results
end

function _OUT.SetHDMIOptions(mode,colorspace,audio_sampling)
	local changed
	if mode then
		_OUT.output_config.hdmi_mode = mode
		changed = true
	end
	if colorspace then
		_OUT.output_config.colorspace = colorspace
		changed = true
	end
	if audio_sampling then
		_OUT.output_config.hdmi_audio_sampling = audio_sampling
		changed = true
	end
	if changed then
		_OUT.output_config:__TRIGGER(true)
	end

	local i,v
	for i,v in ipairs(_OUT.output_config.bind_interfaces) do
		local vodrv = vo_drivers.GetInterfaceDriver(v)
		local en_auto_hdmi
		if vodrv and vodrv.interface_type == "HDMI" then
			local aud_out = require "mod_AudioOut"
			aud_out.SetOutputFormat( "HDMI", { sampling=_OUT.output_config.hdmi_audio_sampling, channels=2 } )
			if _OUT.output_config.hdmi_mode == "DVI" or _OUT.output_config.hdmi_mode == "dvi" then
				vodrv:SetFailDefaultMode("DVI")
				vodrv:SetVideoColorOptions("RGB444")
				vodrv:EnableAudio(false)
			elseif _OUT.output_config.hdmi_mode == "" or _OUT.output_config.hdmi_mode == "AUTO" or _OUT.output_config.hdmi_mode == "auto" then
				vodrv:SetFailDefaultMode("DVI")
				vodrv:SetVideoColorOptions(_OUT.output_config.colorspace)
				vodrv:EnableAudio(true)
				vodrv:SetAudioFormat(_OUT.output_config.hdmi_audio_sampling, 2, 16 )
				en_auto_hdmi = true
			else --HDMI
				vodrv:SetFailDefaultMode("HDMI")
				vodrv:SetVideoColorOptions(_OUT.output_config.colorspace)
				vodrv:EnableAudio(true)
				vodrv:SetAudioFormat(_OUT.output_config.hdmi_audio_sampling, 2, 16 )
			end

			if en_auto_hdmi or _OUT.output_config.colorspace == "AUTO" or _OUT.output_config.colorspace == "auto" then
				vodrv:EnableAutoDetection(true)
			else
				vodrv:EnableAutoDetection(false)
			end
		end
	end
end

function _OUT.SetAudioGain( atype, volume, local_en )
	local aud_out = require "mod_AudioOut"
	local vol = tonumber(volume)
	local only_adjust_vol
	local dis = 0

	if local_en then
		if atype == "analog" then
			_OUT.output_config.analog_audio.local_enabled = tonumber(local_en) or 0
		else
			_OUT.output_config.embedded_audio.local_enabled = tonumber(local_en) or 0
		end
	end

	if vol then
		only_adjust_vol = true
		if vol < -121 then
			vol = -121
		elseif vol > 6 then
			vol = 6
		end

		if atype == "analog" then
			_OUT.output_config.analog_audio.volume = vol
			dis = _OUT.output_config.analog_audio.disabled or 0
		else
			atype = "embedded"
			_OUT.output_config.embedded_audio.volume = vol
			dis = _OUT.output_config.embedded_audio.disabled or 0
		end
	else
		local set_disable
		if volume == "enable" then
			dis = 0
			set_disable = true
		elseif volume == "disable" then
			dis = 1 
			set_disable = true
		end
		--else volume == "none", but can skip this.

		--TODO: Improve this! (analog only allow one output)
		if set_disable and dis == 0 and atype == "analog" then
			local other_name, other_obj
			for other_name, other_obj in pairs(_M.all_outputs) do
				if other_obj ~= _OUT then
					other_obj.SetAudioGain( "analog", "disable" )
				end
			end
		end

		if atype == "analog" then
			if set_disable then
				_OUT.output_config.analog_audio.disabled = dis
			else
				dis = _OUT.output_config.analog_audio.disabled or 0
			end
			vol = _OUT.output_config.analog_audio.volume or 0
		else
			atype = "embedded"
			if set_disable then
				_OUT.output_config.embedded_audio.disabled = dis
			else
				dis = _OUT.output_config.embedded_audio.disabled or 0
			end
			vol = _OUT.output_config.embedded_audio.volume or 0 
		end
	end
	_OUT.output_config:__TRIGGER(true)
	
	--当音频使能enable时,val却设置为0，会导致音量不正常
	if dis == 0 then
		--local pretty = require "pl.pretty"
		--pretty.dump(sys_audio_config.outputs)
		local i,v
		for i,v in ipairs(_OUT.output_config.bind_audio_outputs) do
			if sys_audio_config.outputs and sys_audio_config.outputs[v] and sys_audio_config.outputs[v].audio_type == atype then
				if volume == "enable" then
					aud_out.SetOutputVolume( v, "unmute" )
				else
					aud_out.SetOutputVolume( v, vol )
				end
			end
		end

		if sys_video_config.outputs.common_audio_bind then
			for i,v in ipairs(sys_video_config.outputs.common_audio_bind) do
				if sys_audio_config.outputs and sys_audio_config.outputs[v] and sys_audio_config.outputs[v].audio_type == atype then
					if volume == "enable" then
						aud_out.SetOutputVolume( v, "unmute" )
					else
						aud_out.SetOutputVolume( v, vol )
					end
				end
			end
		end
	end

	if not only_adjust_vol then
		_OUT.UpdateAudioOutputs()
	end

	return true
end

function _OUT.GetAudioGain( atype )
	local dis, vol, local_en
	if atype == "analog" then
		dis = _OUT.output_config.analog_audio.disabled or 0
		vol = _OUT.output_config.analog_audio.volume or 0
		local_en = _OUT.output_config.analog_audio.local_enabled or 0
	else
		atype = "embedded"
		dis = _OUT.output_config.embedded_audio.disabled or 0
		vol = _OUT.output_config.embedded_audio.volume or 0 
		local_en = _OUT.output_config.embedded_audio.local_enabled or 0
	end
	return dis, vol, local_en
end

function _OUT.BindAudioOutput( intfs )
	if type(intfs) ~= "table" then
		intfs = { intfs }
	end
	local i,k,changed
	for i=1, #intfs do
		if _M.audio_output_mapping[ intfs[i] ] then
			local fnd = false
			for k=1, #_OUT.output_config.bind_audio_outputs do
				if _OUT.output_config.bind_audio_outputs[k] == _M.audio_output_mapping[ intfs[i] ] then
					fnd = true
					break
				end
			end
			if not fnd then
				table.insert( _OUT.output_config.bind_audio_outputs, _M.audio_output_mapping[ intfs[i] ] )
				changed = true
			end
		end
	end
	if changed then
		_OUT.output_config:__TRIGGER(true)
		_OUT.UpdateAudioOutputs()
	end
	return true
end

function _OUT.UnbindAudioOutput( intfs )
	if type(intfs) ~= "table" then
		intfs = { intfs }
	end
	local i,k
	local changed
	for i=1, #intfs do
		if _M.audio_output_mapping[ intfs[i] ] then
			for k=1, #_OUT.output_config.bind_audio_outputs do
				if _OUT.output_config.bind_audio_outputs[k] == _M.audio_output_mapping[ intfs[i] ] then
					changed = true
					table.remove( _OUT.output_config.bind_audio_outputs, k )
					break
				end
			end
		end
	end
	if changed then
		_OUT.output_config:__TRIGGER(true)
		_OUT.UpdateAudioOutputs()
	end
	return true
end

function _OUT.UpdateOutputEncMode()
	local i
	local out_binds = {}
	local out_screens = {}

	--outputX  bind HDMI         -> 通过WBC采集输出图像   -> encode
	--outputX  bind SDI(BT1120)  -> 通过VI(0)采集输出图像 -> encode
	if features.DEV_BT1120_VI then
		local o_name, o_obj
		for o_name, o_obj in pairs(_M.all_outputs) do
			if o_obj.screen and o_obj.screen.deviceAttr and type(o_obj.screen.deviceAttr.enIntfType) == "table" then
				out_screens[ o_obj.device ] = o_obj.screen
				for i=1, #o_obj.screen.deviceAttr.enIntfType do
					if o_obj.screen.deviceAttr.enIntfType[i] == "BT1120" then
						out_binds[ o_obj.device ] = "BTx"
						break
					else
						_dbg.Info( "OUTPUT", "++++++++++++++++HDMI use WBC---o_obj.device:", o_obj.device ,"type:", o_obj.screen.deviceAttr.enIntfType[i])
						out_binds[ o_obj.device ] = "WBC"
					end
				end
			end
		end

		if out_binds["hd:0"] == "BTx" then
			out_screens["hd:0"]:ReleaseWbc() 	--hd:0用VI
			if out_screens["hd:1"] then
				out_screens["hd:1"]:CreateWbc() --hd:1用WBC
			end
		elseif out_binds["hd:1"] == "BTx" then
			out_screens["hd:1"]:ReleaseWbc() 	--hd:1用VI
			if out_screens["hd:0"] then
				out_screens["hd:0"]:CreateWbc() --hd:1用WBC
			end
		elseif out_binds["hd:0"] == "WBC" then
			out_screens["hd:1"]:ReleaseWbc() 	--hd:1用VI
			if out_screens["hd:0"] then
				out_screens["hd:0"]:CreateWbc() --hd:0用WBC
			end
		elseif out_binds["hd:1"] == "WBC" then
			out_screens["hd:0"]:ReleaseWbc() 	--hd:0用VI
			if out_screens["hd:1"] then
				out_screens["hd:1"]:CreateWbc() --hd:1用WBC
			end
		else
			if out_screens["hd:1"] then
				out_screens["hd:1"]:ReleaseWbc()
			end
			if out_screens["hd:0"] then
				out_screens["hd:0"]:CreateWbc()
			end
		end

		_dbg.Info( "OUTPUT", "****************************** OUTPUT1=", out_binds["hd:0"], " | OUTPUT2=", out_binds["hd:1"] )
	end


	local mo_encoder = require "mod_Encoder"
	mo_encoder.UpdateEncoders()

end

--NOTE：D300切换SDI和HDMI和IPSTREAM，都要把WBC释放掉，再创建，否则出现bug：Select Hi-encoders fd timeout!
function _OUT.Update_D300_Bind_OutputEncMode(isIpStream)	
	local o_name, o_obj
	if isIpStream then  -- 假如是ipstream
		--WBC释放
		for o_name, o_obj in pairs(_M.all_outputs) do
			if o_obj.screen and o_obj.screen.deviceAttr and type(o_obj.screen.deviceAttr.enIntfType) == "table" then
				_dbg.Info( "OUTPUT", "*****NET******release***", o_obj.device )
				o_obj.screen:ReleaseWbc()
			end
		end
		--WBC创建
		for o_name, o_obj in pairs(_M.all_outputs) do
			if o_obj.screen and o_obj.screen.deviceAttr and type(o_obj.screen.deviceAttr.enIntfType) == "table" then
				if o_obj.device:match("0") and _OUT.name:match("0") then 
					_dbg.Info( "OUTPUT", o_obj.device, "**NET**wbc create" )
					o_obj.screen:CreateWbc() --假如outpu1选择了ipstream则创建output1的wbc
				elseif o_obj.device:match("1") and _OUT.name:match("1") then 
					_dbg.Info( "OUTPUT", o_obj.device, "**NET**wbc create" )
					o_obj.screen:CreateWbc() --假如outpu2选择了ipstream则创建output2的wbc
				end
			end
		end
	else  -- 假如是SDI或HDMI
		--WBC释放
		for o_name, o_obj in pairs(_M.all_outputs) do
			if o_obj.screen and o_obj.screen.deviceAttr and type(o_obj.screen.deviceAttr.enIntfType) == "table" then
				_dbg.Info( "OUTPUT", "****BT1120 HDMI*******wbc release", o_obj.device )
				o_obj.screen:ReleaseWbc()
			end
		end
		--WBC创建
		for o_name, o_obj in pairs(_M.all_outputs) do
			if o_obj.screen and o_obj.screen.deviceAttr and type(o_obj.screen.deviceAttr.enIntfType) == "table" then
				-- _dbg.Info( "OUTPUT", "-------------------------------",o_obj.device)
				-- local pt = require "pl.pretty"
				-- pt.dump(o_obj.screen.deviceAttr)
				-- _dbg.Info( "OUTPUT", "**BT1120 HDMI**OUTPUT**************id", id, ",o_obj.device",o_obj.device,"_OUT.name",_OUT.name)
				for i=1, #o_obj.screen.deviceAttr.enIntfType do
					--_dbg.Info( "OUTPUT", " -----add hhy----- enIntfType[i]", o_obj.screen.deviceAttr.enIntfType[i] )
					if o_obj.screen.deviceAttr.enIntfType[i] == "NET" or o_obj.screen.deviceAttr.enIntfType[i] == "VGA" then
						_dbg.Info( "OUTPUT", "****BT1120 HDMI*******wbc create",o_obj.device )
						o_obj.screen:CreateWbc() --哪个outpu选择了ipstream则创建哪个output的wbc
						break
					end
				end
			end
		end
	end 
	
end

function _OUT.Update_D300_Unbind_OutputEncMode(isIpStream)
	local o_name, o_obj
	if isIpStream then  -- 假如是ipstream
		--NOTE 不处理
	else  -- 假如是SDI或HDMI
		--WBC释放
		for o_name, o_obj in pairs(_M.all_outputs) do
			if o_obj.screen and o_obj.screen.deviceAttr and type(o_obj.screen.deviceAttr.enIntfType) == "table" then
				_dbg.Info( "OUTPUT", "****BT1120 HDMI*******wbc release", o_obj.device )
				o_obj.screen:ReleaseWbc()
			end
		end
		--WBC创建
		for o_name, o_obj in pairs(_M.all_outputs) do
			if o_obj.screen and o_obj.screen.deviceAttr and type(o_obj.screen.deviceAttr.enIntfType) == "table" then
				-- _dbg.Info( "OUTPUT", "-------------------------------",o_obj.device)
				-- local pt = require "pl.pretty"
				-- pt.dump(o_obj.screen.deviceAttr)
				-- _dbg.Info( "OUTPUT", "**BT1120 HDMI**OUTPUT**************id", id, ",o_obj.device",o_obj.device,"_OUT.name",_OUT.name)
				for i=1, #o_obj.screen.deviceAttr.enIntfType do
					--_dbg.Info( "OUTPUT", " -----add hhy----- enIntfType[i]", o_obj.screen.deviceAttr.enIntfType[i] )
					if o_obj.screen.deviceAttr.enIntfType[i] == "NET" or o_obj.screen.deviceAttr.enIntfType[i] == "VGA" then
						_dbg.Info( "OUTPUT", "****BT1120 HDMI*******wbc create",o_obj.device )
						o_obj.screen:CreateWbc() --哪个outpu选择了ipstream则创建哪个output的wbc
						break
					end
				end
			end
		end
	end 
end


local function checkStreamerUsing()
	local streamer = require "mod_Streamer"
	local k, set
	for k, set in pairs(streamer.config) do
		if string.sub(set.source_stream,1,string.len("OUTPUT")) == "OUTPUT" then
			return true
		end
	end
	return false
end

--NOTE 删除没选择的OutputStream，解决点击ipstream在:流服务->源选择 选项显示混乱问题
function _OUT.RemoveNoSelectedOutputStreams(isSelect)
	local mo_encoder = require "mod_Encoder"
	if _OUT.name == "hd0" then  --NOTE 假如点击的是output1则删除hd1的output2
		for id, obj in pairs(mo_encoder.local_out_encoders) do
			if isSelect then
				if id and (id == "OUTPUT2" or id == "OUTPUT2:sub") then
					if streams.GetStreamObject(id) ~= nil then
						_dbg.Info( "OUTPUT", "..RemoveOutputStream ID:",id )
						streams.RemoveOutputStream(id)
					end	
				end	
			else
				if id and (id == "OUTPUT2" or id == "OUTPUT2:sub" or id == "OUTPUT1" or id == "OUTPUT1:sub") then
					if streams.GetStreamObject(id) ~= nil then
						_dbg.Info( "OUTPUT", "..RemoveOutputStream ID:",id )
						streams.RemoveOutputStream(id)
					end	
				end	
			end 
		end
	elseif _OUT.name == "hd1" then --NOTE 假如点击的是output2则删除hd1的output1
		for id, obj in pairs(mo_encoder.local_out_encoders) do
			if isSelect then
				if id and (id == "OUTPUT1" or id == "OUTPUT1:sub") then
					if streams.GetStreamObject(id) ~= nil then
						_dbg.Info( "OUTPUT", "..RemoveOutputStream ID:",id )
						streams.RemoveOutputStream(id)
					end	
				end	
			else 
				if id and (id == "OUTPUT2" or id == "OUTPUT2:sub" or id == "OUTPUT1" or id == "OUTPUT1:sub") then
					if streams.GetStreamObject(id) ~= nil then
						_dbg.Info( "OUTPUT", "..RemoveOutputStream ID:",id )
						streams.RemoveOutputStream(id)
					end	
				end	
			end 
		end
	end
end

function _OUT.BindVideoInterface( intfs )
	_dbg.Info( "OUTPUT", "------BindVideoInterface start..." )

	--NOTE 检测当前ipstream是否正在使用，正在使用则禁止切换ipstream
	if type(intfs) == "string" and intfs == "NET" and checkStreamerUsing() then
		_dbg.Info( "OUTPUT", "The output streamer is using..." )
		--return false, "The output streamer is using..."
		return false, "setOutputStreamUsedError"
	end

	if _OUT.screen then
		--TODO: process the special 'NET'/'CVBS' output interface
		if type(intfs) ~= "table" then
			intfs = { intfs }
		end

		local i,k,changed
		local isSelectIpStream = false

		local other_name, other_obj
		for other_name, other_obj in pairs(_M.all_outputs) do
			--print(other_obj)
			if other_obj ~= _OUT then
				other_obj.UnbindVideoInterface( intfs, {no_encoder_update=true} )	--Remove other's binding.
			end
		end

		for i=1, #intfs do
			if intfs[i] == "NET" then
				isSelectIpStream = true
			end

			if not _M.enabled_interfaces[ intfs[i] ] then
				_dbg.Warn( "OUTPUT", _OUT.name, ": Unsupported or not enabled video port of:", intfs[i] )
			else
				if _OUT.screen:BindInterface( intfs[i] ) then
					local fnd = false
					for k=1, #_OUT.output_config.bind_interfaces do
						if _OUT.output_config.bind_interfaces[k] == intfs[i] then
							fnd = true
							break
						end
					end
					if not fnd then
						table.insert( _OUT.output_config.bind_interfaces, intfs[i] )
						changed = true
					end
				end
			end
		end

		if changed then
			_OUT.output_config:__TRIGGER(true)

			_OUT.Start() --Try start but if already started, no impact.
			 
			_OUT.screen:ChangeResolution(_OUT.output_config.resolution)

			_OUT.SetOutputLayout()
			_OUT.SetHDMIOptions()
			_OUT.BindAudioOutput( intfs ) --Bind posible audio mixer for corresponding video interface.

			_dbg.Info( "OUTPUT", "Output video interface binding is changed, update overlay and output encoding...")
			--if #_OUT.output_config.bind_interfaces > 0 then
			OSDChangeResolution(_OUT.name,_OUT.screen.syncDetail.width,_OUT.screen.syncDetail.height)
			--end

			--Update the output encoding working mode.
			if features.DEVICE_TYPE == "E_DOCTOR" or features.DEVICE_TYPE == "E_DOCTOR_S" then
				_OUT.UpdateOutputEncMode() 
			end
		end

		-- --NOTE 解决点击ipstream在:流服务->源选择 选项显示混乱问题
		if isSelectIpStream and features.FEATURE_IPStream then
			_OUT.RemoveNoSelectedOutputStreams(true)
		end

		--NOTE D300等系列产品
		if changed and features.FEATURE_IPStream then
			_OUT.Update_D300_Bind_OutputEncMode(isSelectIpStream) ----NOTE解决D300的出现bug：Select Hi-encoders fd timeout!
		end

		return true
	else
		return false, "Invalid screen object"
	end
end

function _OUT.UnbindVideoInterface( intfs, action_flags )
--NOTE: action_flags is used internal. It's a table, with:
--{
--   no_osd_update = true     #Do not update OSD.
--   no_encoder_update = true #Do not update encoders.
--}

	--NOTE 检测当前ipstream是否正在使用，正在使用则禁止关闭ipstream
	if type(intfs) == "string" and intfs == "NET" and checkStreamerUsing() then
		_dbg.Info( "OUTPUT", "The output streamer is using..." )
		--return false, "The output streamer is using..."
		return false, "setOutputStreamUsedError"
	end

	if type(action_flags) ~= "table" then
		action_flags = {}
	end
	if _OUT.screen then
		if type(intfs) ~= "table" then
			intfs = { intfs }
		end
		local i,k
		local changed
		local isSelectIpStream = false

		for i=1, #intfs do
			if intfs[i] == "NET" then
				isSelectIpStream = true
			end

			if _OUT.screen:UnbindInterface( intfs[i] ) then
				for k=1, #_OUT.output_config.bind_interfaces do
					if _OUT.output_config.bind_interfaces[k] == intfs[i] then
						changed = true
						table.remove( _OUT.output_config.bind_interfaces, k )
						break
					end
				end
			end
		end
		
		if changed then
			_OUT.output_config:__TRIGGER(true)

			_OUT.UnbindAudioOutput( intfs ) --Unbind posible audio mixer for corresponding video interface.
			_OUT.screen:ChangeResolution(_OUT.output_config.resolution)

			_OUT.SetOutputLayout() --Relayout, if no interfaces bound, remove all channels.

			if not action_flags.no_osd_update then
				_dbg.Info( "OUTPUT", "Output video interface binding is changed, update overlay...")
				--if #_OUT.output_config.bind_interfaces > 0 then
					OSDChangeResolution(_OUT.name,_OUT.screen.syncDetail.width,_OUT.screen.syncDetail.height)
				--end
			end
			
			if features.DEVICE_TYPE == "E_DOCTOR" or features.DEVICE_TYPE == "E_DOCTOR_S" then
				if not action_flags.no_encoder_update  then
					_dbg.Info( "OUTPUT", "Output video interface binding is changed, update output encoders...")
					--Update the output encoding working mode.
					_OUT.UpdateOutputEncMode()  
				end
			end
		end

		--NOTE 解决点击ipstream在:流服务->源选择 选项显示混乱问题 D300等系列产品
		if isSelectIpStream and features.FEATURE_IPStream then
			_OUT.RemoveNoSelectedOutputStreams(false)
		end

		if changed and features.FEATURE_IPStream then --D300等系列产品
			_OUT.Update_D300_Unbind_OutputEncMode(isSelectIpStream) ----NOTE解决D300的出现bug：Select Hi-encoders fd timeout!
		end

		-- local pt = require "pl.pretty"
		-- pt.dump(streams)

		return true
	else
		return false, "Invalid screen object"
	end
end

function _OUT.CleanupNoUsedStreams()
	local item, _nu
	for item, _nu in pairs(_OUT.deprecated_streams) do
		local using = false
		local i, v
		for i,v in pairs(_OUT.layout_map) do
			if _OUT.output_config.channels[i] and _OUT.output_config.channels[i].stream == item then
				using = true
				break
			end
		end

		if not using then
			local streamObj = streams.GetStreamObject( item )
			if streamObj then
				_dbg.Info( "OUTPUT", "Close the stream output because it is not used for display now! Stream-id:", item )
				streamObj:Stop(_OUT.name, "output")
			end

			local subStreamObj = streams.GetStreamObject( item .. ":sub" )
			if subStreamObj then
				_dbg.Info( "OUTPUT", "Close the sub stream priview because it is not used for priview now! Stream-id:", item , ":sub")
				subStreamObj:Stop(_OUT.name, "preview")
			end
		end
	end

	_OUT.deprecated_streams = {}
	collectgarbage()
end

function _OUT.GetOutputLayout()
	if not _OUT.output_config then
		return nil
	end

	local layout_id = _OUT.output_config.layout

	if #_OUT.output_config.bind_interfaces == 0 then
		return nil
	end

	local layout_table = layout.GetLayoutDetail( layout_id )
	if not layout_table then
		return nil
	end
	return layout_table
end

function _OUT.SetBorder()
	if not _OUT.output_config then
		_dbg.Error( "OUTPUT", "Fail to _OUT.output_config is nil" )
		return
	end

	local border = _OUT.output_config.border
	-- pt.dump(border)
	if border and _OUT.screen then
		_OUT.screen:SetVpssBorder(border.enabled, border.bordercolor or 0xffffff)
	end	
end

function _OUT.SetOutputLayout(layout_id)
	local old_layout_id = _OUT.output_config.layout
	if layout_id then
		local layout_table = layout.GetLayoutDetail( layout_id )
		if layout_table then
			_OUT.output_config.layout = layout_id
			_OUT.output_config:__TRIGGER(true)
		end
	end

	if not _OUT.screen or not _OUT.screen.initialized then
		_dbg.Error( "OUTPUT", "Fail to layout output since screen object is not created/initialized." )
		return false
	end

	layout_id = layout_id or _OUT.output_config.layout
	
	--No interfaces bound, so drop anything.
	if #_OUT.output_config.bind_interfaces == 0 then
		layout_id = -1
	end

	if not _OUT.layout_created or layout_id ~= old_layout_id then
		local max_wnds = 0
		local i
		if layout_id == -1 then
			_OUT.screen:ReLayout(true) --First, disable all channels for updating.
		else
			local layout_table = layout.GetLayoutDetail( layout_id )
			if not layout_table then
				_dbg.Error( "OUTPUT", "Invalid (unsupported) layout mode:", layout_id )
				return false
			end

			_OUT.screen:ReLayout(true) --First, disable all channels for updating.

			max_wnds = #layout_table
			--Destroy no used windows
			for i=1, max_wnds do
				local detail = _OUT.layout_map[i]
				if not detail then
					--Create the new window.
					detail = {
						voChannel_id = i --mast modify: start 1
					}
					
					_OUT.layout_map[i] = detail
				end
				--NOTE: If channel is already created, then update the layout configure.
				_OUT.screen:CreateChannel( detail.voChannel_id, layout_table[i] )
			end
			
			local layoutobj = layout.GetLayout( layout_id )
			if layoutobj and layoutobj.img_id then
				if layoutobj.img_id ~= -1 and layoutobj.img_id ~= 0 then
					_dbg.Info( "OUTPUT", "Change BackGround img_id", layoutobj.img_id  )
					if not _M.OSDSetBackGround(true, _OUT.name, layoutobj.img_id) then
						_dbg.Error( "OUTPUT", "OSDSetBackGround error!!!"  )
					end
				else
					_dbg.Info( "OUTPUT", "Do not change BackGround..." )
				end
			end
		end

		--Remove no used windows
		for i=max_wnds+1, sys_config.MAX_SUPPORT_WINDOWS do
			local detail = _OUT.layout_map[i]
			if detail then
				_OUT.screen:RemoveChannel( detail.voChannel_id )
				_OUT.layout_map[i] = nil
				if _OUT.output_config.channels[i] and _OUT.output_config.channels[i].stream then
					_OUT.deprecated_streams[ _OUT.output_config.channels[i].stream ] = true
					streams.SetStreamAudioOutputs( _OUT.output_config.channels[i].stream, _OUT.name, {} ) --Clear audio mixers
				end
			end
		end
	
		_OUT.screen:ReLayout()
		_OUT.UpdateStreamBinding()
		if layout_id == -1 then
			_OUT.layout_created = false
		else
			_OUT.layout_created = true
		end
	end

	-- _OUT.InitBackGround()
	--_OUT.SetBackGroundOSD()
	-- mast 20201201
	coaf.ScheduleDelayedTask(function()
		_OUT.UpdatePicInfo()
	end, 0)

	return true
end

function _OUT.GetSupportLayouts()
	return layout.EnumLayouts()
end

function _OUT.BindStream( wnd_id, stream_id )
	_dbg.Info( "OUTPUT", "BindStream wnd_id=",wnd_id,"stream_id=",stream_id)
	if wnd_id > 0 and wnd_id <= sys_config.MAX_SUPPORT_WINDOWS then
		if not _OUT.output_config.channels[wnd_id] and stream_id and stream_id ~= "" then
			_OUT.output_config.channels[wnd_id] = {
				stream = stream_id
			}
			_OUT.output_config:__TRIGGER(true)
			_OUT.UpdateStreamBinding()
		elseif _OUT.output_config.channels[wnd_id] and _OUT.output_config.channels[wnd_id].stream ~= stream_id then
			if _OUT.output_config.channels[wnd_id].stream and _OUT.output_config.channels[wnd_id].stream ~= "" then
				_OUT.deprecated_streams[ _OUT.output_config.channels[wnd_id].stream ] = true
				streams.SetStreamAudioOutputs( _OUT.output_config.channels[wnd_id].stream, _OUT.name, {} ) --Release previous stream mixers
			end

			if not stream_id or stream_id == "" then
				_OUT.output_config.channels[wnd_id] = nil --Drop it
			else
				_OUT.output_config.channels[wnd_id].stream = stream_id
			end

			_OUT.output_config:__TRIGGER(true)
			_OUT.UpdateStreamBinding()
		end
		return true
	else
		_dbg.Error( "OUTPUT", "Binding stream: windows ID:", wnd_id, "out of range!" )
		return false
	end
end

function _OUT.RemoveStreamByWindow( wnd_id )
	return _OUT.BindStream( wnd_id, nil )
end

function _OUT.RemoveStream( stream_id )
	local k,v
	for k,v in pairs(_OUT.output_config.channels) do
		if v.stream == stream_id then
			_OUT.output_config.channels[k] = nil
			_OUT.output_config:__TRIGGER(true)
		end
	end
	_OUT.output_config:__TRIGGER(true)
	if stream_id and stream_id ~= "" then
		_OUT.deprecated_streams[ stream_id ] = true
		streams.SetStreamAudioOutputs( stream_id, _OUT.name, {} ) --Release previous audio mixers.
		_OUT.dispatcher_config[ stream_id ] = nil
		_M.audio_out_config:__TRIGGER(true)
	end
	_OUT.UpdateStreamBinding()
	return true
end

function _OUT.UpdateStreamBinding()
	_dbg.Info( "OUTPUT", "UpdateStreamBinding") 
	local k, v
	for k, v in pairs(_OUT.layout_map) do
		if _OUT.output_config.channels[k] and _OUT.output_config.channels[k].stream then
			local streamObj = streams.GetStreamObject( _OUT.output_config.channels[k].stream )
			if not streamObj then
				_dbg.Info( "OUTPUT", "Window:", k, ", previous bound stream of", _OUT.output_config.channels[k].stream, "is not exist! Drop it!" )
				_OUT.screen:BindChannelSource( v.voChannel_id, nil, true )
			else
				_dbg.Info( "OUTPUT", "streamObj:Start", _OUT.name, "output")
				streamObj:Start(_OUT.name, "output") --Enable stream's output
				_OUT.screen:BindChannelSource( v.voChannel_id, streamObj:GetDisplayVpss(_OUT.name), true )
			end

			local subStreamObj = streams.GetStreamObject( _OUT.output_config.channels[k].stream .. ":sub" )
			if subStreamObj then
				_dbg.Info( "OUTPUT", "subStreamObj:Start", _OUT.name, "preview")
				subStreamObj:Start(_OUT.name, "preview") --Enable stream's output
			end
		else
			_dbg.Info( "OUTPUT", "Window:", k, ", no stream bound so drop output!" )
			_OUT.screen:BindChannelSource( v.voChannel_id, nil, true )
		end
	end

	_OUT.SetBorder()
	_OUT.UpdateAudioOutputs()

	-- if not _OUT.bg_vpsschn then
	-- 	_OUT.bg_vpsschn = _OUT.bg_vpss_grp:GetChannel(0)
	-- 	if not _OUT.bg_vpsschn then
	-- 		_dbg.Error( "OUTPUT", "Fail to get bg_vpsschn, so bg OSD may be wrong!" )
	-- 		return
	-- 	end
	-- end
	-- _OUT.screen:BindChannelSource( 0, _OUT.bg_vpsschn, true )
end

function _OUT.UpdateAudioOutputs()
	local i,v
	local audio_outs = {}
	local common_ifs = {}
	if sys_video_config.outputs.common_audio_bind then
		for i,v in ipairs(sys_video_config.outputs.common_audio_bind) do
			local atype = sys_audio_config.outputs[v] and sys_audio_config.outputs[v].audio_type
			if atype == "analog" then
				if _OUT.output_config.analog_audio.disabled == 0 then
					table.insert( audio_outs, v )
					common_ifs[v] = true
				end
			else
				if _OUT.output_config.embedded_audio.disabled == 0 then
					table.insert( audio_outs, v )
					common_ifs[v] = true
				end
			end
		end
	end

	for i,v in ipairs(_OUT.output_config.bind_audio_outputs) do
		local atype = sys_audio_config.outputs[v] and sys_audio_config.outputs[v].audio_type
		if atype == "analog" then
			if _OUT.output_config.analog_audio.disabled == 0 and not common_ifs[v] then
				table.insert( audio_outs, v )
			end
		else
			if _OUT.output_config.embedded_audio.disabled == 0 and not common_ifs[v] then
				table.insert( audio_outs, v )
			end
		end
	end

	if _OUT.virtual_mixer then
		if type(_OUT.virtual_mixer) == "table" then
			for i, v in ipairs(_OUT.virtual_mixer) do
				table.insert(audio_outs, v)
			end
		elseif type(_OUT.virtual_mixer) == "string" then
			table.insert( audio_outs, _OUT.virtual_mixer)
		end
	end

	local funcs = {}
	local mod_AudioOut = require "mod_AudioOut"
	for i, v in pairs(_OUT.layout_map) do
		if _OUT.output_config.channels[i] and _OUT.output_config.channels[i].stream then
			streams.SetStreamAudioOutputs( _OUT.output_config.channels[i].stream, _OUT.name, audio_outs )
			table.insert(funcs, function()
				local stream_id = _OUT.output_config.channels[i].stream
				local dispatcher = mod_AudioOut.GetChannelSink(stream_id)
				if not _OUT.dispatcher_config[stream_id] then
					local zero = dispatcher:SetGainRange(200, -51, 20)
					local volume = 142
					dispatcher:SetGainValue(zero, zero)
					local detail = streams.GetStreamDetail(stream_id)
					local mute = 0
					if detail.audio_enables then
						local enable = detail.audio_enables[_OUT.name]
						mute = enable and 0 or 1
						dispatcher:Mute(not enable)
					end
					_OUT.dispatcher_config[stream_id] = {zero = zero, volume = volume, mute = mute}
				else
					dispatcher:SetGainRange(200, -51, 20)
					local volume = _OUT.dispatcher_config[stream_id].volume
					local mute = _OUT.dispatcher_config[stream_id].mute
					if mute == 1 then
						dispatcher:Mute(true)
					else
						dispatcher:Mute(false)
						dispatcher:SetGainValue(volume, volume)
					end
				end
				_M.audio_out_config:__TRIGGER(true)
			end)
		end
	end

	coaf.ScheduleDelayedTask(function()
		for _, f in ipairs(funcs) do
			f()
		end
	end, 0)
end

function _OUT.EnableAudio( wnd_id, en )
	wnd_id = tostring(wnd_id)
	if not wnd_id then
		wnd_id = "*"
	end
	local i,v
	for i, v in pairs(_OUT.output_config.channels) do
		if wnd_id == "*" or wnd_id == tostring(i) then
			local stream_id = _OUT.output_config.channels[i] and _OUT.output_config.channels[i].stream or nil
			if stream_id and stream_id ~= "" then
				streams.ChangeStreamDetail( stream_id, { audio_enables = { [_OUT.name] = en } } )
			end
		end
	end
	return true
end

function _OUT.ModifyAudioVolume(id, volume, mute)
	mute = tonumber(mute)
	id = tonumber(id)
	local stream_id = _OUT.output_config.channels[id] and _OUT.output_config.channels[id].stream
	if not stream_id then
		return false, "Invalid id"
	end

--设置媒体源的声音
--参数为channel id号，即 MANUAL.5
	local mod_AudioOut = require "mod_AudioOut"
	if not mod_AudioOut then
		return false, "Not mixer"
	end

	local dispatcher = mod_AudioOut.GetChannelSink(stream_id)
	if not dispatcher then
		return false, "Not Dispatcher"
	end

	_OUT.dispatcher_config[stream_id].mute = mute
	if tonumber(volume) then
		_OUT.dispatcher_config[stream_id].volume = volume
	end

	_M.audio_out_config:__TRIGGER(true)
	if mute ~= 1 then
		dispatcher:SetGainRange(200, -51, 20)
		dispatcher:Mute(false)
		if tonumber(volume) then
			dispatcher:SetGainValue(volume, volume)
		else
			dispatcher:SetGainValue(_OUT.dispatcher_config[stream_id].volume, _OUT.dispatcher_config[stream_id].volume)
		end
	else
		dispatcher:Mute(true)
	end
	return true
end

function _OUT.GetAudioVolume()
	local info = {}
	for i, v in pairs(_OUT.layout_map) do
		local stream_id = _OUT.output_config.channels[i] and _OUT.output_config.channels[i].stream or nil
		if stream_id and _OUT.dispatcher_config[stream_id] then
			local detail = streams.GetStreamDetail(stream_id)
			table.insert(info, {
				name = detail.name,
				volume = _OUT.dispatcher_config[stream_id].volume,
				mute = _OUT.dispatcher_config[stream_id].mute,
				stream_id = stream_id,
				idx = i,
			})
		end
	end
	return info
end

function _OUT.Update()
	_OUT.CleanupNoUsedStreams()
end

function _OUT.GetShowLayouts()
	if ( not _OUT.output_config ) then 
		return { "1" }
	end
	local ret = _OUT.output_config.show_layouts
	if not ret or #ret == 0 then
		if _OUT.device:match( "sd:" ) then
			return { "1" }
		elseif _OUT.device == "hd:1" then
			return sys_video_config.outputs.hd1_default_layout
		else
			return sys_video_config.outputs.hd0_default_layout
		end
	else
		return ret
	end
end

function _OUT.ChangeShowLayouts(items)
	if type(items) == "table" then
		_OUT.output_config.show_layouts = items
		_OUT.output_config:__TRIGGER(true)
		return true
	else
		return false
	end
end

function _OUT.EnableBorder(items)
	if type(items) == "table" then
		_OUT.output_config.border.enabled = items.enabled or 0
		_OUT.output_config.border.bordercolor = items.bordercolor or 0xffffff
		_OUT.output_config:__TRIGGER(true)
		_OUT.SetBorder()
		return true
	else
		return false
	end
end

function _OUT.GetInterfacesBinding()
	local rslt = {}
	local i
	for i=1, #_OUT.enabled_interfaces do
		local bind = false
		local j
		for j=1, #_OUT.output_config.bind_interfaces do
			if _OUT.output_config.bind_interfaces[j] == _OUT.enabled_interfaces[i] then
				bind = true
				break
			end
		end
		table.insert( rslt, { id=_OUT.enabled_interfaces[i], name=_OUT.enabled_interfaces[ _OUT.enabled_interfaces[i] ], enable = bind and 1 or 0, bind = bind } )
	end
	return rslt
end

function _OUT.GetOutputLayoutID()
	if not _OUT.output_config then
		return nil
	end

	local layout_id = _OUT.output_config.layout
	local layout_table = layout.GetLayoutDetail( layout_id )
	if not layout_table then
		return nil
	end
	return layout_id
end

_M.all_outputs[ output_device ] = _OUT
return _OUT

--==================================================
--} end of the wrapper
--==================================================
end

function _M.UpdateStreamBinding()
	local k,obj
	for k,obj in pairs(_M.all_outputs) do
		obj.UpdateStreamBinding()
	end
end

function _M.Update()
	local k,obj
	for k,obj in pairs(_M.all_outputs) do
		obj.Update()
	end
end

-- new add for Streams RemoveAllNetworkStream()
function _M.RemoveStream( stream_id )
	local k,obj
	for k,obj in pairs(_M.all_outputs) do
		obj.RemoveStream( stream_id )
	end
end

function _M.GetOutputByName(out_name)
	local k,obj
	for k,obj in pairs(_M.all_outputs) do
		if obj.name == out_name then
			return obj
		end
	end
	return nil
end

return _M

