local sys_config = require "sys_config"
local mod_Decoder = require "mod_Decoder"
local coaf = require "coaf.core"
local _dbg = require "coaf.debug"

local decoder = require "soc.decoder"
local aud_out = require "mod_AudioOut"
local sip_cfg = require "configs.sip_config"
local sip_event = require "sip.event"
local gen_account = require "sip.general_account"
local gb_account = require "sip.gb28181_account"

require "kmCore"

local account_adapters = {
	["sip"] = gen_account,
	["gb28181"] = gb_account
}

local _M = {
	config = nil,
	sipEp = nil,
	sipEp_timer = nil,
	all_accounts = {},
	local_binds = {},
	account_base_id = 1
}

local function parse_addr( addr )
	local tok_addr, tok_port = addr:match("%s*([^%s]*)%s*:%s*(%d*)")
	if tok_addr then
		if tok_addr == '' then
			tok_addr = '*'
		end
		return '*', tonumber(tok_port) or 5060
	else
		if tonumber(addr) then
			return '*', tonumber(addr)
		else
			return addr, 5060
		end
	end
end

local function parse_binds( binds )
	if type(binds) ~= "string" then
		_dbg.Error( "SIP", "Invalid binding addresses!" )
		return {}
	else
		local rslt = {}
		local used_ports = {}
		binds = binds .. ","
		local x
		for x in binds:gmatch("(.-)[,;]+") do
			if x ~= "" then
				local addr,port = parse_addr(x)
				if not used_ports[port] then
					used_ports[port] = addr .. ":" .. port
					table.insert( rslt, {address=addr,port=port,bind=addr .. ":" .. port} )
				else
					_dbg.Warn( "SIP", "Binding address of '" .. x .. "', the port", port, "is already used by previous binding", used_ports[port], "! So current address will be skipped." )
				end
			end
		end
		return rslt
	end
end

-----------------------------------------
--Sip module polling
-----------------------------------------

local sip_is_checking = false
local function sip_check_timer()
	if sip_is_checking then
		return
	end
	sip_is_checking = true

	if _M.sipEp then
		_M.sipEp:Polling(0)
	end

	gen_account.Step()
	--TODO: gb_account.Step()
	
	sip_is_checking = false
end

-----------------------------------------

function _M.Init()
	_M.config = sip_cfg.Load()
	if not _M.config then
		_dbg.Error( "SIP", "Invalid configure file!" )
		_M.config = {
			binds = "*:5060",
			accounts = {},
			__TRIGGER = function()
			end
		}
	end
	local base_id
	local k,v
	for k,v in pairs(_M.config.accounts) do
		if not base_id or k > base_id then
			base_id = k
		end
	end
	_M.account_base_id = base_id or 0
end

function _M.FreeInstance()
	if _M.sipEp_timer then
		_M.sipEp_timer:stop()
		_M.sipEp_timer = nil
		sip_is_checking = false
	end
	if _M.sipEp then
		_M.FreeAccount() --Remove all accounts.
		_M.sipEp = nil
		_M.local_binds = {}
		collectgarbage()
	end
end

function _M.CreateInstance()
	if _M.sipEp then
		_dbg.Warn( "SIP", "CreateInstance: the SIP instance is already created!" )
		return true
	end

	_M.local_binds = parse_binds( _M.config.binds )
	if #_M.local_binds == 0 then
		_dbg.Error( "SIP", "CreateInstance: No binding address configured, fail!" )
		return false
	end
	table.sort( _M.local_binds, function(a,b)
		return a.port < b.port
	end )

	local binds = {}
	local k,v
	for k,v in ipairs(_M.local_binds) do
		table.insert( binds, v.bind )
	end

	_M.sipEp = KM_STREAMING.SIPEndpoint.instance( binds, "REMUX", 3, 60)
	if not _M.sipEp then
		_dbg.Error( "SIP", "CreateInstance: Fail to create SIP endpoint instance!" )
		_M.local_binds = {}
		return false
	end

	sip_event.Init(_M.sipEp)
	sip_is_checking = false
	_M.sipEp_timer = coaf.NewTimer( sip_check_timer, 0.01, 0.1 )
	if _M.sipEp_timer then
		_M.sipEp_timer:start()
	end
	return true
end

function _M.SetLocalBinds(binds)
	if _M.config.binds ~= binds then
		_M.config.binds = binds
		_M.config:__TRIGGER(true)

		local binds = parse_binds( _M.config.binds )
		table.sort( binds, function(a,b)
			return a.port < b.port
		end)

		local changed
		if #binds == #_M.local_binds then
			local i
			for i=1, #binds do
				if binds[i].bind ~= _M.local_binds[i].bind then
					changed = true
					break
				end
			end
		else
			changed = true
		end

		if changed then
			_M.FreeInstance()
			_M.CreateInstance()
			_M.SetupAccount()
		end
	end
end

function _M.GetLocalBinds()
	return _M.config.binds, _M.local_binds
end

function _M.GetAccountList()
	local rslt = {}
	local k,v
	for k,v in pairs(_M.config.accounts) do
		table.insert( rslt, {
			id = k,
			name = v.name,
			type = v.protocol or "sip",
			created = (_M.all_accounts[k] ~= nil)
		})
	end

	table.sort( rslt, function(a,b)
		return a.id < b.id
	end)
	return rslt
end

function _M.GetAccountStatus(id)
	if id then
		local source_id
		if _M.config.accounts[id] then
			source_id = _M.config.accounts[id].source_id
		else
			source_id = ""
		end

		if _M.all_accounts[id] then
			local acc_obj = _M.all_accounts[id].acc_obj
			local reg_status, reg_code = acc_obj:GetRegStatus()
			local call_status, in_count, out_count = acc_obj:GetCallStatus()
			return {
				account_status = acc_obj:GetAccountStatus(),
				register_status = reg_status,
				register_code = reg_code,
				call_status = call_status,
				call_in_count = in_count,
				call_out_count = out_count,
				source_id = source_id
			}
		else
			return { account_status = "not_created",
				register_status = "none",
				register_code = 0,
				call_status = "none",
				call_in_count = 0,
				call_out_count = 0,
				source_id = source_id
			}
		end
	else
		return { account_status = "invalid",
			register_status = "none",
			register_code = 0,
			call_status = "none",
			call_in_count = 0,
			call_out_count = 0,
			source_id = ""
		}
	end
end

function _M.GetAccountOptions(id)
	return _M.config.accounts[id]
end

function _M.SetupAccount(id) --If without id, setup all accounts.
	if not _M.sipEp then
		_dbg.Error( "SIP", "SetupAccount: invalid SIP endpoint instance, not create or fail?" )
		return false
	end

	if id then
		if not _M.config.accounts[id] then
			_dbg.Error( "SIP", "SetupAccount: Invalid account id:", id )
			return false
		end
		
		if _M.all_accounts[id] then
			_dbg.Info( "SIP", "SetupAccount: Modify account settings of id:", id )
			local opts = _M.config.accounts[id]
			if not opts or not opts.localAccount or opts.localAccount == "" then
				_dbg.Info( "SIP", "SetupAccount: Invalid account options or account is empty of:", id, "remove the account!")
				return _M.FreeAccount(id)
			else
				return _M.all_accounts[id].acc_obj:Setup(opts)
			end
		else
			local opts = _M.config.accounts[id]
			if not opts or not opts.localAccount or opts.localAccount == "" then
				_dbg.Info( "SIP", "SetupAccount: Invalid account options or account is empty of:", id, "account is not created!")
				return false
			end

			local protocol = "sip"
			if opts.protocol == "GB" or opts.protocol == "GB28181" or
			   opts.protocol == "gb" or opts.protocol == "gb28181" then
				--protocol = "gb28181"
				--TODO: Now not support gb yet.
				protocol = "sip"
			end
			_dbg.Info( "SIP", "SetupAccount: Creating new account:", id, " with protocol", protocol )
			local acc_adapt = account_adapters[protocol]
			local acc_obj = acc_adapt.CreateAccount( id, _M.sipEp, opts )
			if not acc_obj then
				_dbg.Error( "SIP", "SetupAccount: Fail to create account:", id )
				return false
			else
				_dbg.Info( "SIP", "SetupAccount: account:", id, "created!" )
			end
			
			_M.all_accounts[id] = {
				protocol = protocol,
				config = _M.config.accounts[id],
				acc_obj = acc_obj
			}

			return true
		end
	else
		local id, cfg
		for id,cfg in pairs(_M.config.accounts) do
			_M.SetupAccount(id)
		end
		return true
	end
end

function _M.FreeAccount(id)
	if id then
		local obj = _M.all_accounts[id]
		if not obj then
			_dbg.Warn( "SIP", "Free account:", id, "but it's not created or invalid!" )
			return false
		end

		_dbg.Info( "SIP", "FreeAccount: (", id, ") Destroying..." )
		obj.acc_obj:Destroy()
		_M.all_accounts[id] = nil
		_dbg.Info( "SIP", "FreeAccount: (", id, ") Destroied." )
		collectgarbage()
		return true
	else
		local k,v
		for k,v in pairs(_M.all_accounts) do
			_dbg.Info( "SIP", "FreeAccount: (all) Destroying account of ID=", k )
			if v.acc_obj then
				v.acc_obj:Destroy()
				v.acc_obj = nil
			end
		end
		_M.all_accounts = {}
		_dbg.Info( "SIP", "FreeAccount: (all) Destroied." )
		collectgarbage()
		return true
	end
end

function _M.RemoveAccount(id)
	if id then
		_dbg.Info( "SIP", "RemoveAccount: account id=", id )
		_M.FreeAccount(id)
		_M.config.accounts[id] = nil
		_M.config:__TRIGGER(true)
		return true
	else
		_dbg.Warn( "SIP", "RemoveAccount: Invalid account id for removing!" )
		return false
	end
end

function _M.AddAccount(opts)
	if type(opts) ~= "table" then
		_dbg.Error( "SIP", "AddAccount: Invalid account options!" )
		return false
	else
		_M.account_base_id = _M.account_base_id + 1
		local id = _M.account_base_id
		_dbg.Info( "SIP", "AddAccount: New account id=", id, ", account name is:", opts.name )
		_M.config.accounts[id] = opts
		_M.config:__TRIGGER(true)
		_M.SetupAccount(id)
		return true
	end
end

function _M.ModifyAccount(id, opts)
	if id and type(opts) == "table" then
		if _M.config.accounts[id] then
			local old_opts = _M.config.accounts[id]
			local k,v
			local changed
			for k,v in pairs(opts) do
				if k == "gb28181_options" and type(v) == "table" then
					local kk,vv
					for kk,vv in pairs(v) do
						if old_opts.gb28181_options[kk] ~= vv then
							changed = true
							old_opts.gb28181_options[kk] = vv
						end
					end
				elseif k == "call_in_options" and type(v) == "table" then
					local kk,vv
					for kk,vv in pairs(v) do
						if old_opts.call_in_options[kk] ~= vv then
							changed = true
							old_opts.call_in_options[kk] = vv
						end
					end
				elseif k == "call_out_options" and type(v) == "table" then
					local kk,vv
					for kk,vv in pairs(v) do
						if old_opts.call_out_options[kk] ~= vv then
							changed = true
							old_opts.call_out_options[kk] = vv
						end
					end
				else
					if old_opts[k] ~= v then
						changed = true
						old_opts[k] = v
					end
				end
			end
			if changed then
				_M.config:__TRIGGER(true)
				_M.SetupAccount(id)
				return true
			else
				return true
			end
		else
			_dbg.Warn( "SIP", "ModifyAccount: specified account id:", id, "is not exist!" )
			return false
		end
	else
		_dbg.Warn( "SIP", "ModifyAccount: Invalid id or invalid account options!" )
		return false
	end
end

function _M.MakeCall( stream_id )
	local k,obj
	for k,obj in pairs(_M.all_accounts) do
		if obj.acc_obj and obj.acc_obj:IsOwnedStream(stream_id) then
			return obj.acc_obj:MakeCall(stream_id)
		end
	end
	return false
end

function _M.AnswerCall( stream_id )
	local k,obj
	for k,obj in pairs(_M.all_accounts) do
		if obj.acc_obj and obj.acc_obj:IsOwnedStream(stream_id) then
			return obj.acc_obj:AnswerCall(stream_id)
		end
	end
	return false
end

function _M.HangupCall( stream_id )
	local k,obj
	for k,obj in pairs(_M.all_accounts) do
		if obj.acc_obj and obj.acc_obj:IsOwnedStream(stream_id) then
			return obj.acc_obj:HangupCall(stream_id)
		end
	end
	return false
end

function _M.SelectSource( stream_id, src_stream_id )
	local streamObj = _M.GetStreamObject( stream_id )
	if streamObj then
		_dbg.Info( "SIP", "Stream:", stream_id, "selecting source with:", src_stream_id )
		streamObj:SelectSource( src_stream_id )
	end
end

function _M.GetAccountObject( acc_id )
	local k,obj
	for k,obj in pairs(_M.all_accounts) do
		if obj.acc_obj and obj.acc_obj:MatchSipAccountID(acc_id) then
			return obj.acc_obj
		end
	end
	return nil
end

function _M.UpdateAllSourceBindings()
	local k,obj
	for k,obj in pairs(_M.all_accounts) do
		if obj.acc_obj then
			obj.acc_obj:UpdateAllSourceBindings()
		end
	end
end

function _M.GetStreamObject( stream_id )
	local k,obj
	local sobj
	for k,obj in pairs(_M.all_accounts) do
		if obj.acc_obj then
			sobj = obj.acc_obj:GetStreamObject( stream_id )
			if sobj then
				return sobj
			end
		end
	end
	return nil
end

return _M

