local coaf = require "coaf.core"
local _dbg = require "coaf.debug"
local sip_stream = require "sip.stream"
local mod_Streams = require "mod_Streams"

require "kmCore"

local _M = {
	all_accounts = {}
}
--------------------------------------------
local acc_meta = {}
--------------------------------------------

function acc_meta:Init()
	self:SetupAccount()
	self:UpdateDefaultCall()
end

function acc_meta:Destroy()
	self:ClearAllCalls()
	if self.sipAccID then
		_dbg.Info( "SIP-ACC", "Destroy: remove sip account", self.sipAccID )
		self.account_status = "not_created"
		self.sipEp:removeAccount( self.sipAccID )
		self.sipAccID = nil
	end
	_M.all_accounts[ self.id ] = nil
end

function acc_meta:Setup(opts)
	self.options = opts
	if self.current_account_detail ~= self:__generate_account_detail() then
		self:SetupAccount()
	end
	self:UpdateDefaultCall()
	return true
end

function acc_meta:PendingHangupCall(call_id, call_obj, delay_count)
	self.hangup_pendings[call_id] = {
		call_obj = call_obj,
		delay = delay_count
	}
end

function acc_meta:ClearHangupPending(call_id)
	self.hangup_pendings[call_id] = nil
end

function acc_meta:GetCallByCallID(call_id)
	local k,v
	for k,v in ipairs(self.all_calls) do
		if v.call_id == call_id then
			return v
		end
	end
	return nil
end

function acc_meta:UpdateDefaultCall(call_id, call_status)
	if not self.sipAccID then
		return nil
	end

	local default_call
	if #self.all_calls == 0 then
		default_call = {
			id = "SIP." .. "defaul-" .. self.id,
			dir = "out",
			is_default = true,
			call_id = call_id,
			status = call_status or "idle",
			name = self.options.name
		}
		self.all_calls[ default_call.id ] = default_call
		self.all_calls[1] = default_call
	else
		default_call = self.all_calls[1]
		default_call.name = self.options.name
		if call_id and call_id ~= default_call.call_id then
			default_call.call_id = call_id
			if default_call.stream_obj then
				default_call.stream_obj:ChangeCallID( self.sipAccID, call_id)
			end
		end
		if call_status then
			default_call.status = call_status
		end
	end

	if self.options.call_out_options and self.options.call_out_options.target and self.options.call_out_options.target ~= "" and
		self.options.call_out_options.create_stream then
		default_call.need_in_streams = true
	else
		default_call.need_in_streams = false
	end
	
	if default_call.status == "idle" or default_call.status == "error" then
		if default_call.stream_obj and not default_call.need_in_streams then
			_dbg.Info( "SIP-ACC", "Remove default call stream from stream list:", default_call.id )
			--Schedule later.
			--coaf.ScheduleDelayedTask( function()
			mod_Streams.RemoveSipStream( default_call.id )
			default_call.stream_obj = nil
			--end, 0 )
		elseif default_call.need_in_streams and not default_call.stream_obj then
			_dbg.Info( "SIP-ACC", "Creating default call stream object of:", self.options.localAccount, "... , status = in_streams" )
			default_call.stream_obj = sip_stream.CreateStream( self.sipEp, default_call.id, default_call.name, default_call, self.sipAccID, default_call.call_id )
			if not default_call.stream_obj then
				_dbg.Error( "SIP-ACC", "Fail to create default call stream object for", self.options.localAccount )
				return nil
			end
			default_call.stream_obj:SetBufferOptions( self.options.buffer_option )
			_dbg.Info( "SIP-ACC", "Add default call stream to stream list:", default_call.id )
			--Schedule later.
			coaf.ScheduleDelayedTask( function()
			mod_Streams.AddSipStream( default_call.id, default_call.name )
			end, 0 )
		end
	else
		if not default_call.stream_obj then
			_dbg.Info( "SIP-ACC", "Creating new default call stream object of:", self.options.localAccount, "..." )
			default_call.stream_obj = sip_stream.CreateStream( self.sipEp, default_call.id, default_call.name, default_call, self.sipAccID, default_call.call_id )
			if not default_call.stream_obj then
				_dbg.Error( "SIP-ACC", "Fail to create default call stream object for", self.options.localAccount )
				return nil
			end
			default_call.stream_obj:SetBufferOptions( self.options.buffer_option )

			_dbg.Info( "SIP-ACC", "Add new default call stream to stream list:", default_call.id )
			--Schedule later.
			--mod_Streams.AddSipStream( default_call.id, default_call.name )
			coaf.ScheduleDelayedTask( function()
			mod_Streams.AddSipStream( default_call.id, default_call.name )
			end, 0 )
		end
	end

	return default_call
end

function acc_meta:ClearAllCalls()
	local k,call_obj
	for k,call_obj in ipairs(self.all_calls) do
		if call_obj.call_id then
			if call_obj.stream_obj then
				call_obj.stream_obj:DisableStream()
			end
			self:ClearHangupPending( call_obj.call_id )
			--Hangup directly.
			self.sipEp:hangupCall( call_obj.call_id, 0 )
			call_obj.call_id = nil
		end
		if call_obj.stream_obj then
			mod_Streams.RemoveSipStream( call_obj.id )
			--It should destroy stream after RemoveSipStream
			call_obj.stream_obj = nil
		end
	end
	self.all_calls = {}
end

function acc_meta:ClearCall( call_obj )
	if not call_obj then
		return
	end

	if call_obj.stream_obj then
		call_obj.stream_obj:DisableStream()
	end

	if call_obj.call_id then
		self:ClearHangupPending( call_obj.call_id )
		--Hangup directly.
		self.sipEp:hangupCall( call_obj.call_id, 0 )
		call_obj.call_id = nil
	end

	if not call_obj.is_default then
		if call_obj.stream_obj then
			mod_Streams.RemoveSipStream( call_obj.id )
			--It should destroy stream after RemoveSipStream
			call_obj.stream_obj = nil
		end
		self.all_calls[ call_obj.id ] = nil
		local k,v
		for k,v in ipairs(self.all_calls) do
			if v == call_obj then
				table.remove( self.all_calls, k )
				break
			end
		end
	else
		call_obj.status = "idle"
		call_obj.dir = "out"
		call_obj.peer = ""
		self:UpdateDefaultCall()
	end
end

function acc_meta:__generate_account_detail()
	local rslt = ""
	if self.options.localAccount and self.options.localAccount ~= "" then
		rslt = rslt .. "AC=" .. self.options.localAccount
	end
	rslt = rslt .. "|"
	if self.options.registerServer and self.options.registerServer ~= "" then
		rslt = rslt .. "SVR=" .. self.options.registerServer .. ":" .. (self.options.registerPort or 5060)
	end
	rslt = rslt .. "|REALM=" .. (self.options.realm and self.options.realm ~= "" and self.options.realm or "*")
	if self.options.localAddress then
		rslt = rslt .. "|LA=" .. self.options.localAddress
	end
	rslt = rslt .. "|U=" .. (self.options.username or "") .. "|P=" .. (self.options.password or "")
	return rslt
end

function acc_meta:SetupAccount()
	if self.options.localAccount and self.options.localAccount ~= "" then
		self:ClearAllCalls()
		local acc = self.options.localAccount
		local reg
		if self.options.registerServer and self.options.registerServer ~= "" then
			reg = "sip:" .. self.options.registerServer .. ":" .. (self.options.registerPort or 5060)
		end

		if self.sipAccID then
			self.sipEp:modifyAccount( self.sipAccID, {
				id = acc,
				register = reg,
				realm = self.options.realm and self.options.realm ~= "" and self.options.realm or "*",
				username = self.options.username,
				password = self.options.password,
				address = self.options.localAddress,
				registerNow = false
			})

			self.current_account_detail = self:__generate_account_detail()
			if reg then
				self.sipEp:registerAccount( self.sipAccID )
			end
			self:UpdateDefaultCall()
			return true
		else
			self.sipAccID = self.sipEp:addAccount( acc, {
				register = reg,
				realm = self.options.realm and self.options.realm ~= "" and self.options.realm or "*",
				username = self.options.username,
				password = self.options.password,
				address = self.options.localAddress,
				registerNow = false
			})
			if not self.sipAccID then
				self.current_account_detail = nil
				_dbg.Error( "SIP-ACC", "SetupAccount: fail to create account for", acc )
				self.account_status = "error"
			else
				self.current_account_detail = self:__generate_account_detail()
				_dbg.Info( "SIP-ACC", "SetupAccount: Added SIP account:", acc )
				self.account_status = "created"
				if reg then
					self.sipEp:registerAccount( self.sipAccID )
				end
				self:UpdateDefaultCall()
			end
			return true
		end
	else
		self.current_account_detail = nil
		if self.sipAccID then
			_dbg.Info( "SIP-ACC", "SetupAccount: remove sip account", self.sipAccID, "since no local account configured" )
			self:ClearAllCalls()
			self.account_status = "not_created"
			self.sipEp:removeAccount( self.sipAccID )
			self.sipAccID = nil
		end
		return true
	end
end

function acc_meta:GetRegStatus()
	if self.account_status ~= "created" then
		return "idle", 0
	else
		return self.register_status, self.register_code
	end
end

function acc_meta:GetAccountStatus()
	return self.account_status
end

function acc_meta:GetCallStatus(justState)
	local call_status = "idle"
	local i,v
	local in_count = 0
	local out_count = 0
	for i,v in ipairs(self.all_calls) do
		if v.call_id then
			call_status = "calling"
			if justState then
				break
			end
			if v.dir == "in" then
				in_count = in_count + 1
			else
				out_count = out_count + 1
			end
		end
	end
	return call_status, in_count, out_count
end

function acc_meta:MakeCall()
	if not self.sipAccID then
		_dbg.Error( "SIP-ACC", self.options.localAccount, "Making call but current account is not created!" )
		return false, "invalid"
	end

	local call_status = self:GetCallStatus(true)
	if call_status == "calling" then
		_dbg.Error( "SIP-ACC", self.options.localAccount, "Making call but current account call status is:", call_status, ", call fail!" )
		return false, "busy"
	end

	local default_call = self.all_calls[1]
	if not default_call then
		self:UpdateDefaultCall()
		default_call = self.all_calls[1]
	end
	if not default_call then
		_dbg.Error( "SIP-ACC", self.options.localAccount, "Making call but fail to setup default call!" )
		return false, "fail"
	end

	if not self.options.call_out_options or
		not self.options.call_out_options.target or
		self.options.call_out_options.target == "" then
		_dbg.Error( "SIP-ACC", self.options.localAccount, "Making call but with invalid target, call fail!" )
		return false, "bad-target"
	end

	_dbg.Info( "SIP-ACC", self.options.localAccount, "Calling target:", self.options.call_out_options.target )
	--NOTE: status first.
	default_call.status = "ring"
	default_call.dir = "out"
	default_call.peer = self.options.call_out_options.target

	local call_id = self.sipEp:makeCall( self.sipAccID, self.options.call_out_options.target )
	_dbg.Info( "SIP-ACC", "Make call result (call-id) is:", call_id )
	if not call_id or call_id < 0 then
		_dbg.Error( "SIP-ACC", self.options.localAccount, "Fail to make call to target:", self.options.call_out_options.target )
		default_call.call_id = nil
		default_call.status = "error"
		return false, "fail"
	else
		self:UpdateDefaultCall(call_id)
		return true
	end
end

function acc_meta:AnswerCall(stream_id)
	if not self.sipAccID then
		_dbg.Error( "SIP-ACC", self.options.localAccount, "Answer call but current account is not created!" )
		return false, "invalid"
	end

	local current_call = self.all_calls[stream_id]
	if not current_call or not current_call.call_id then
		_dbg.Error( "SIP-ACC", self.options.localAccount, "Answer call but specified invalid stream-id or stream without valid call-id" )
		return false, "invalid"
	end

	--NOTE: MUST *FIRST* change the status, since the callbacks may change status while answering.
	current_call.status = "answer"
	local st = self.sipEp:answerCall( current_call.call_id, 200 )
	if st ~= 0 then
		_dbg.Error( "SIP-ACC", self.options.localAccount, "Answer call", stream_id, "(call:", current_call.call_id, ") fail!" )
		current_call.status = "error"
		--current_call.call_id = nil
		self:PendingHangupCall( current_call.call_id, current_call )
		return false, "fail"
	end
	return true
end

function acc_meta:IsOwnedStream(stream_id)
	if not stream_id then
		return false
	elseif self.all_calls[stream_id] ~= nil then
		return true
	else
		return false
	end
end

function acc_meta:HangupCall(stream_id)
	if not self.sipAccID then
		_dbg.Error( "SIP-ACC", self.options.localAccount, "Hangup call but current account is not created!" )
		return false, "invalid"
	end

	local current_call = self.all_calls[stream_id]
	if not current_call or not current_call.call_id then
		_dbg.Error( "SIP-ACC", self.options.localAccount, "Hangup call but specified invalid stream-id or stream without valid call-id" )
		return false, "invalid"
	end

	--NOTE: status first.
	current_call.status = "hangup"
	self:PendingHangupCall( current_call.call_id, current_call )
	return true
end

function acc_meta:OnIncomingCall(call_id, args)
	local from = args and args.from or "UNKNOWN"
	if not self.options.call_in_options.allow then
		_dbg.Info( "SIP-ACC", "Incoming call:", call_id, "from:", from, "but not allowed, hangup it!" )
		self:PendingHangupCall( call_id, nil )
		return
	end

	local peer_name, from_acc = from:match('"([^"]+)"%s*<(.+)>')
	if not peer_name then
		peer_name, from_acc = from:match("'([^']+)'%s*<(.+)>")
	end

	if not peer_name then
		peer_name = from
		from_acc = from:match('<(.+)>')
		if not from_acc then
			from_acc = from
		end
	end

	local call_status = self:GetCallStatus(true)
	local current_call
	if call_status == "idle" then
		current_call = self:UpdateDefaultCall( call_id, "ring" )
		if not current_call or not current_call.stream_obj then
			_dbg.Error( "SIP-ACC", "Incoming call:", call_id, "from:", from, "but cannot setup default call context!" )
			self:PendingHangupCall( call_id, nil )
			return
		end

		current_call.dir = "in"
		current_call.peer = peer_name
	elseif self.options.call_in_options.policy == 1 then
		_dbg.Info( "SIP-ACC", "Incoming call:", call_id, "from:", from, "but only one call allowed, hangup it!" )
		self:PendingHangupCall( call_id, nil )
		return
	elseif self.options.call_in_options.policy == 2 then --Replace call.
		local k,v
		for k,v in ipairs(self.all_calls) do
			if v.call_id and v.dir == "in" then --Replace old incoming call.
				current_call = v
				break
			end
		end

		if current_call then
			_dbg.Info( "SIP-ACC", "Replace old call of id=", current_call.call_id, ", peer=", current_call.peer, " with new call:", call_id, ", from:", from )
			--NOTE: status first.
			current_call.status = "hangup"
			--NOTE: Don't pass current call object to pending queue.
			self:PendingHangupCall( current_call.call_id, nil, 30 ) --Delay 30 loops.(~=3seconds)
		end
	end

	if not current_call then
		_dbg.Info( "SIP-ACC", "Incoming call:", call_id, "from:", from, "creating call context..." )
		current_call = {
			id = "SIP." .. "from=" .. from_acc .. "-" .. self.id,
			dir = "in",
			call_id = call_id,
			status = "ring",
			name = self.options.name .. "@" .. call_id,
			peer = peer_name,
			need_in_streams = true
		}
		current_call.stream_obj = sip_stream.CreateStream( self.sipEp, current_call.id, current_call.name, current_call, self.sipAccID, call_id )
		if not current_call.stream_obj then
			_dbg.Error( "SIP-ACC", "Fail to create stream object for incoming call:", call_id, "from:", from )
			self:PendingHangupCall( call_id, nil )
			return
		end
		current_call.stream_obj:SetBufferOptions( self.options.buffer_option )
		self.all_calls[ current_call.id ] = current_call
		table.insert( self.all_calls, current_call )
	else
		current_call.dir = "in"
		current_call.call_id = call_id
		current_call.status = "ring"
		current_call.name = self.options.name .. "@" .. call_id
		current_call.peer = peer_name
	end

	if current_call.stream_obj then
		current_call.stream_obj:ChangeCallID( self.sipAccID, call_id )
		coaf.ScheduleDelayedTask( function()
		mod_Streams.AddSipStream( current_call.id, current_call.name )
		end, 0 )
	end

	if self.options.call_in_options.auto_answer then
		current_call.status = "answer"
		local tag = self.sipEp:answerCall( call_id, 200 )
		if tag ~= 0 then
			_dbg.Error( "SIP-ACC", "Auto answer call of:", call_id, "from:", from, "fail! Hangup the call!" )
			--NOTE: status first
			current_call.status = "error"
			self:PendingHangupCall( call_id, current_call )
		end
	end
end

function acc_meta:OnCallState(call_id, state)
    local call = self:GetCallByCallID(call_id)
	if call then 
		if state == "DISCONNCTD" then
			_dbg.Info( "SIP-ACC", "Call:", call_id, "is terminated." )
			call.call_id = nil
			self:ClearHangupPending( call_id )
			self:ClearCall( call )
		elseif state == "CONFIRMED" then
			_dbg.Info( "SIP-ACC", "Call:", call_id, "is established." )
			call.stream_obj:EnableStream()
			call.status = "calling"
		else
			_dbg.Info( "SIP-ACC", "Call:", call_id, "State:", state )
		end                                                                                                                                         
	end
end

function acc_meta:OnRegisterState(reg_flag, state, code)
	if self.register_status ~= "registered" or state ~= "registering" then
		self.register_status = state      
		self.register_code = code        
	end

	if reg_flag == "reg" and (state == "error" or state == "registered") then
		self.register_retry = os.time()  --retry register.
	end
end

--此处需要开启 否则更新会话无法播放
function acc_meta:OnUpdateVideoStream( call_id, video_port, audio_port )
    local call = self:GetCallByCallID(call_id)
	if call and call.stream_obj then 
		call.stream_obj:SelectSource( self.options.source_id )
		call.stream_obj:EnableStream()
	end
end

function acc_meta:OnStopVideoStream( acc_id, call_id )
    local call = self:GetCallByCallID(call_id)
	if call and call.stream_obj then 
		call.stream_obj:DisableStream()
	else
		--Clear who is delayed using current call's media source, because current call is die.
		local i,v
		for i,v in ipairs(self.all_calls) do
			if v.stream_obj then
				v.stream_obj:ClearMediaSourceUsing( acc_id, call_id )
			end
		end
	end
end

function acc_meta:OnStopAudioStream( acc_id, call_id )
    local call = self:GetCallByCallID(call_id)
	if call and call.stream_obj then 
		call.stream_obj:DisableStream()
	else
		--Clear who is delayed using current call's media source, because current call is die.
		local i,v
		for i,v in ipairs(self.all_calls) do
			if v.stream_obj then
				v.stream_obj:ClearMediaSourceUsing( acc_id, call_id )
			end
		end
	end
end

function acc_meta:OnFixupSDP( sdp_acker )
	--No action for general sip account.
end

function acc_meta:GetStreamObject(stream_id)
	if stream_id then
		if self.all_calls[stream_id] then
			return self.all_calls[stream_id].stream_obj
		else
			return nil
		end
	else
		return nil
	end
end

function acc_meta:UpdateAllSourceBindings()
	local k,obj
	for k,obj in ipairs( self.all_calls) do
		if obj.stream_obj then
			obj.stream_obj:SelectSource()
		end
	end
end

function acc_meta:MatchSipAccountID(acc_id)
	return self.sipAccID == acc_id
end

function acc_meta:Step()
	local k, obj

	for k, obj in pairs(self.hangup_pendings) do
		if obj.delay and obj.delay > 0 then
			obj.delay = obj.delay - 1
		else
			_dbg.Info( "SIP-ACC", "Background hanging up call of:", k )
			self.sipEp:hangupCall( k, 0 )
			--TODO:
			self.hangup_pendings[k] = nil	
		end
	end
	--self.hangup_pendings = {}

	--Update register.
	if self.sipAccID then
		local now = os.time()

		if self.register_retry then
			local diff = now - self.register_retry
			local timeout = 5
			if self.register_status == "registered" then
				timeout = 60 --60 secs re-register.
			end
			if diff < 0 or diff >= timeout then
				self.register_retry = nil
				self.sipEp:registerAccount( self.sipAccID )
			end
		end

		if self.options.call_out_options.auto_call and
			self.options.call_out_options.target and
			self.options.call_out_options.target ~= "" and
			( not self.options.registerServer or self.options.registerServer == "" or self.register_status == "registered")
		then
			local call_state = self:GetCallStatus(true)
			if call_state == "idle" and ( not self.lastMakeCallTime or (now - self.lastMakeCallTime < 0) or (now - self.lastMakeCallTime >= 5) ) then
				self.lastMakeCallTime = now
				self:MakeCall()
			end
		end
	end
	
end

acc_meta.__index = acc_meta

--------------------------------------------

function _M.CreateAccount(id, sipEp, opts)
	if not id then
		_dbg.Error( "SIP-ACC", "Invalid account ID specified!" )
		return nil
	end
	if _M.all_accounts[id] then
		_dbg.Warn( "SIP-ACC", "Create account:", id, "but it is already exist!" )
		return _M.all_accounts[id]
	end

	local acc_obj = {
		id = id,
		sipEp = sipEp,
		sipAccID = nil,
		options = opts,
		current_account_detail = nil,
		account_status = "not_created",
		register_status = "idle",
		register_code = 0,
		all_calls = {},
		hangup_pendings = {}
	}

	setmetatable( acc_obj, acc_meta )
	acc_obj:Init()
	_M.all_accounts[id] = acc_obj
	return acc_obj
end

function _M.Step()
	local k,obj
	for k,obj in pairs(_M.all_accounts) do
		obj:Step()
	end
end

return _M

