require "himpp"
local _dbg = require "coaf.debug"

local SYS = HiMPP.SYS
local AI = HiMPP.AI
local ACODEC = HiMPP.ACODEC

local _M = {
	MAX_CHANNELS = 1,
	aDev = "/dev/acodec",
	aCodec = nil,
	channels = {}
}

function _M.Init(resetCodec)
	if _M.aCodec then
		_dbg.Warn( "AIN", "The audio in module is already initialized.")
		return
	end

	_M.aCodec = ACODEC.CreateInnerCodec( _M.aDev )
	if not _M.aCodec then
		_dbg.Error( "AIN", "Fail to create audio Codec!" )
		return
	end

	if _M.aCodec:Open() < 0 then
		_dbg.Error( "AIN", "Fail to open audio Codec!" )
		_M.aCodec = nil
		return
	end

	if resetCodec then
		_M.aCodec:Reset()
	end

	_M.aCodec:Close()
end

function _M.Cleanup()
	local k, chanObj
	for k, chanObj in pairs(_M.channels) do
		if chanObj then
			chanObj:Destroy()
		end
	end
	_M.channels = {}

	if _M.aCodec then
		_M.aCodec:Open()
		_M.aCodec:Reset()
		_M.aCodec:Close()
		_M.aCodec = nil
	end
end

--------------------------------------------------

local channel_meta = {}

-- Options:
-- {
-- 		channel = "MONO"/"STEREO",
-- 		sampleRate = ...,
-- 		sampleWidth = 8/16/24,
-- 		sampleSize = 256/512/...
-- }

function channel_meta:SetAudioOption( opts )
	local changed = false
	local srChanged = false
	if opts.channel and opts.channel ~= self.aiPubAttr.enSoundmode then
		self.aiPubAttr.enSoundmode = opts.channel
		changed = true
	end

	if opts.sampleRate and opts.sampleRate ~= self.aiPubAttr.enSamplerate then
		self.aiPubAttr.enSamplerate = opts.sampleRate
		changed = true
		srChanged = true
	end

	if opts.sampleWidth and opts.sampleWidth ~= self.aiPubAttr.enBitwidth then
		self.aiPubAttr.enBitwidth = opts.sampleWidth
		changed = true
	end

	if opts.sampleSize and opts.sampleSize ~= self.aiPubAttr.u32PtNumPerFrm then
		self.aiPubAttr.u32PtNumPerFrm = opts.sampleSize
		changed = true
	end

	if srChanged then
		_M.aCodec:Open()
		_M.aCodec:SetSampleRate( self.aiPubAttr.enSamplerate )
		_M.aCodec:Close()
	end

	if changed and self.initialized then
		if self.started then
			self.aiChannel:Disable()
		end
		self.aiDev:Disable()

		local r, err = self.aiDev:SetPubAttr( self.aiPubAttr )
		if not r then
			_dbg.Error( "AIN", "Fail to configure audio IN device, error:", err )
		end

		self.aiDev:Enable()
		if self.started then
			self.aiChannel:Enable()
		end
	end
end

function channel_meta:Init()
	if self.initialized then
		return
	end
	_M.aCodec:Open()
	if _M.aCodec:SelectSource( self.aiSource ) ~= 0 then
		_dbg.Warn( "AIN", "Fail to select audio source:", self.aiSource )
	end

	if _M.aCodec:SetSampleRate( self.aiPubAttr.enSamplerate ) ~= 0 then
		_dbg.Warn( "AIN", "Fail to set audio Codec chip sample rate to ", self.aiPubAttr.enSamplerate )
	end

	_M.aCodec:Close()

	local r, err = self.aiDev:SetPubAttr( self.aiPubAttr )
	if not r then
		_dbg.Error( "AIN", "Fail to configure audio IN device, error:", err )
		return
	end

	r = self.aiDev:Enable()
	if r ~= 0 then
		_dbg.Error( "AIN", "Fail to enable audio IN device! error code:", string.format("%08X",r) )
		return
	end

	self.aiChannel = self.aiDev:GetChannel(0)
	self.initialized = true
end

function channel_meta:Destroy()
	if self.initialized then
		self.aiChannel:Disable()
		self.aiChannel = nil
		self.aiDev:Disable()
		self.initialized = false
	end
end

function channel_meta:Start()
	if not self.initialized then
		self:Init()
	end

	if self.initialized then
		self.aiChannel:Enable()
		self.started = true
		local r, err = self.aiChannel:SetParam( {u32UsrFrmDepth=30} )
		if not r then
			_dbg.Error( "AIN", err )
		end
	end
end

function channel_meta:Stop()
	if self.initialized then
		self.started = false
		self.aiChannel:Disable()
	end
end

function channel_meta:SelectSource(src)
	if src ~= self.aiSource then
		self.aiSource = src
		_M.aCodec:Open()
		if 0 ~= _M.aCodec:SelectSource( self.aiSource ) then
			_dbg.Error( "AIN", "Fail to select audio source:", src )
		end
		_M.aCodec:Close()
	end
end

function channel_meta:SetGain( gain )
	_M.aCodec:Open()
	--[[
	if 0 ~= _M.aCodec:SetInputVolume(gain) then
		_dbg.Error( "AIN", "Fail to set audio input gain to:", gain )
	end
	]]
	--_M.aCodec:SetADCGain( gain, gain )
	_M.aCodec:SetMICGain( gain, gain ) -- This is analog gain, so I think it's best.
	_M.aCodec:Close()
end

function channel_meta:Mute()
	_M.aCodec:Open()
	_M.aCodec:SetMICMute( true, true )
	_M.aCodec:Close()
end

function channel_meta:Unmute()
	_M.aCodec:Open()
	_M.aCodec:SetMICMute( false, false )
	_M.aCodec:Close()
end

channel_meta.__index = channel_meta

---------------------------------------------------

function _M.CreateChannel(id)
	if id < 1 or id > _M.MAX_CHANNELS then
		_dbg.Error( "AIN", "Now audio in only support", _M.MAX_CHANNELS, "channels, but you request create channel ", id )
		return nil
	end

	if _M.channels[id] then
		return _M.channels[id]
	end

	local chanObj = {
		id = id,
		aiDev = AI.GetDevice(id-1),
		aiChannel = nil,
		started = false,

		aiSource = "LINEIN",

		aiPubAttr = {
			enSamplerate = 48000,
			enBitwidth = 16,
			enWorkmode = "I2S_MASTER",
			enSoundmode = "STEREO",
			u32EXFlag = 0,
			u32FrmNum = 30,
			u32PtNumPerFrm = 1024,
			u32ClkChnCnt = 2,
			u32ChnCnt = 2,
			u32ClkSel = 0
		}
	}

	local r = setmetatable( chanObj, channel_meta )
	_M.channels[id] = r

	return r
end

---------------------------------------------------

return _M


