local coaf = require "coaf.core"
local _dbg = require "coaf.debug"
local sys_video_config = require "sys_video_config"

------------------

local _M = {
}

------------------
local VODRV

------------------
function _M.Init(vodrv)
	VODRV = vodrv
end

function _M.FetchBestFormat(res, devType, devId, intfs)
	devType = devType or "hd"
	devId = devId or 0
	--TODO: improve this
	if not res then
		if devType == "hd" then
			res = "1080P60"
		else
			res = "PAL"
		end
	end

	local sup = _M.CheckSupportedFormats(devType,devId,intfs)
	if not sup then --If not any supported format for current interfaces
	--Just return the specified resolution sync-details.
		return _M.GetSyncDetail(res)
	end

	local sync_res,details,t1001 = _M.GetSyncDetail(res, sup)
	if sync_res then --Matched, that's greate.
		return sync_res,details,t1001
	end

	sync_res = VODRV.specialFormatMap[res]
	if sync_res then
		t1001 = true
	else
		t1001 = false
	end

	local want_w, want_h, want_fr, want_intl
	local item = VODRV.syncDetailMap[res]
	if item then
		want_w = item.width
		want_h = item.height
		want_fr = item.frameRate - (t1001 and 1 or 0)
		want_intl = item.interlace
	elseif devType == "sd" then
		want_w = 720
		want_h = 576
		want_fr = 50
		want_intl = true
	else
		want_w = 1920
		want_h = 1080
		want_fr = 60
		want_intl = false
	end

	res = _M.GetNearestResolution(want_w, want_h, want_fr, want_intl, sup )
	return _M.GetSyncDetail(res, sup)
end

------------------

function _M._ResolutionMatch(w, h, fr, res_details)
	-- For example:
	-- want 1920 1080 40
	--
	-- valid:
	-- 1920 1200 50
	-- 1920 1080 60
	-- 1920 1080 50
	-- 1920 1080 30
	-- 1920 1080 25
	-- 1920 1080 24

	local lastMatch
	local diff_fr
	local k,v

	if fr == 59 then
		fr = 60
	elseif fr == 29 then
		fr = 30
	elseif fr == 23 then
		fr = 24
	end

	for k,v in pairs(res_details) do
		if w == v.width and h == v.height then
			diff_fr = fr - v.frameRate
			if diff_fr >= 0 then --break here
				if diff_fr <= 2 or not lastMatch or lastMatch.width ~= w or lastMatch.height ~= h then
					lastMatch = v
				end
				break
			else
				if not lastMatch or lastMatch.width ~= w or lastMatch.height ~= h then
					lastMatch = v
				else
					--Currently, the frameRate *MUST* be large than fr
					local div_old = lastMatch.frameRate/fr
					local div_new = v.frameRate/fr

					div_old = div_old - math.floor(div_old)
					div_new = div_new - math.floor(div_new)

					if div_old ~= 0 and (div_new == 0 or (diff_fr > (fr - lastMatch.frameRate))) then --more closed to framerate
						lastMatch = v
					end
				end
			end
		elseif w >= v.width and h >= v.height then
			if lastMatch and not lastMatch.no_suggest then --Last has goal resolution match, but framerate is not quite near.
				break
			else --Use this resolution, to continue match framerate
				lastMatch = v
				if not v.no_suggest then
					w = v.width
					h = v.height
				end
			end
		else --Resolution is large than want
			if not lastMatch then --Record the best resolution first.
				lastMatch = v
			--Is same as first recorded resolution, but framerate more matched.
			elseif lastMatch.width == v.width and lastMatch.height == v.height then
				local div_old, div_new
				if fr > lastMatch.frameRate then
					div_old = fr / lastMatch.frameRate
				else
					div_old = lastMatch.frameRate/fr
				end

				if fr > v.frameRate then
					div_new = fr/v.frameRate
				else
					div_new = v.frameRate/fr
				end

				div_old = div_old - math.floor(div_old)
				div_new = div_new - math.floor(div_new)

				diff_fr = math.abs(fr - v.frameRate)
				if div_old ~= 0 and (div_new == 0 or (diff_fr < math.abs(fr - lastMatch.frameRate))) then --more closed to framerate
					lastMatch = v
				end
			elseif lastMatch.width > v.width and lastMatch.height > v.height and w <= v.width and h <= v.height then
				lastMatch = v
			end
		end
	end
	return lastMatch
end

------------------

function _M.CheckSupportedFormats(devType,devId,intfs)
	devType = devType or "hd"
	devId = devId or 0
	if type(intfs) ~= "table" then
		intfs = {intfs}
	end
	local k,v
	local sup = {}
	for k,v in pairs(VODRV.syncDetailMap) do
		if devType == "sd" then
			if v.sd then
				sup[k] = v
			end
		else
			if not v.no_hd then
				sup[k] = v
			end
		end
	end

	local kk,vv
	for k,v in pairs(intfs) do
		local vodrv = VODRV.GetInterfaceDriver(v)
		local sup_4k = VODRV.support_4k and vodrv and vodrv.support_4k or false
		if v == "HDMI" then
			for kk,vv in pairs(sup) do
				if vv.no_hdmi or (not sup_4k and vv.four_k_only) then
					sup[kk] = nil
				end
			end
		elseif v == "BT1120" or v == "SDI" then
			for kk,vv in pairs(sup) do
				if vv.no_sdi or (not sup_4k and vv.four_k_only) then
					sup[kk] = nil
				end
			end
		elseif v =="VGA" then
			for kk,vv in pairs(sup) do
				if vv.no_vga or (not sup_4k and vv.four_k_only) then
					sup[kk] = nil
				end
			end
		end
	end

	return sup
end

------------------

-- Specify a resolution ID, translate to excact matching format ID and sync detail.
function _M.GetSyncDetail(res, sup)
	if not sup then
		sup = VODRV.syncDetailMap
	end

	if res then
		local upr_res = res:upper()
		if upr_res == "480I" then
			res = "NTSC"
		elseif upr_res == "576I" then
			res = "PAL"
		elseif upr_res == "1080P" then
			res = "1080P60"
		elseif upr_res == "720P" then
			res = "720P60"
		elseif upr_res == "1080I" then
			res = "1080I50"
		elseif upr_res == "480P" then
			res = "480P60"
		elseif upr_res == "576P" then
			res = "576P50"
		elseif not res:match("%d+x%d+") then
			res = upr_res
		end

		local actual_res = VODRV.specialFormatMap[res]
		local t1001 = false
		if actual_res then
			res = actual_res
			t1001 = true
		end

		local item = sup[ res ]
		if not item then
			return nil
		else
			return res, item, t1001
		end
	else
		return nil
	end
end

--Giving width(w)/height(h)/frame-rate(fr)/interlaced(intl), match a
--nearest resolution format ID.

------------------

function _M.GetNearestResolution(w, h, fr, intl, sup)
	if not sup then
		sup = VODRV.syncDetailMap
	end

	if fr then
		fr = math.floor(fr)
		if fr <= 0 then
			fr = 60
		end
	else
		fr = 60
	end

	local sup_4k = VODRV.support_4k
	local intl_res_details = {}
	local prog_res_details = {}
	local k,v
	for k,v in pairs(sup) do
		if v.interlace then
			if sup_4k or not v.four_k_only then
				table.insert( intl_res_details, {
					width = v.width,
					height = v.height,
					frameRate = v.frameRate,
					no_suggest = v.no_suggest,
					id = k
				} )
			end
		else
			if sup_4k or not v.four_k_only then
				table.insert( prog_res_details, {
					width = v.width,
					height = v.height,
					frameRate = v.frameRate,
					no_suggest = v.no_suggest,
					id = k
				} )
			end
		end
	end

	table.sort( intl_res_details, function(a,b)
		if a.width == b.width and a.height == b.height then
			return a.frameRate > b.frameRate
		elseif a.width == b.width then
			return a.height > b.height
		else
			return a.width > b.width
		end
	end)

	table.sort( prog_res_details, function(a,b)
		if a.width == b.width and a.height == b.height then
			return a.frameRate > b.frameRate
		elseif a.width == b.width then
			return a.height > b.height
		else
			return a.width > b.width
		end
	end)

	local goal
	if intl then --Interlaced format, so check interlaced first.
		goal = _M._ResolutionMatch(w,h,fr,intl_res_details)
		if not goal then
			goal = _M._ResolutionMatch(w,h,fr,prog_res_details)
		end
	else
		goal = _M._ResolutionMatch(w,h,fr,prog_res_details)
		if not goal then
			goal = _M._ResolutionMatch(w,h,fr,intl_res_details)
		end
	end

	if not goal then
		_dbg.Warn( "VO-DRV", "GetNearestResolution: No any valid resolution for w:", w, "h:", h, "fr:", fr, "intl:", intl, "!" )
		local failsafe_res = "720P60"
		if intl then
			if fr == 59 or fr == 29 then
				failsafe_res = "NTSC"
			else
				failsafe_res = "PAL"
			end
		else
			if fr == 59 or fr == 29 then
				failsafe_res = "720P59"
			elseif fr == 50 or (fr >= 23 and fr <= 25) then
				failsafe_res = "720P50"
			else
				failsafe_res = "720P60"
			end
		end
		_dbg.Warn( "VO-DRV", "GetNearestResolution: returns a fail-safe resolution of", failsafe_res )
		return failsafe_res
	else
		--reverseSpecialFormatMap
		if (fr%10) == 9 or (fr%10) == 3 then
			if VODRV.reverseSpecialFormatMap[ goal.id ] then
				return VODRV.reverseSpecialFormatMap[ goal.id ]
			else
				return goal.id
			end
		else
			return goal.id
		end
	end
end

------------------

return _M

