local coaf = require "coaf.core"
local _dbg = require "coaf.debug"
local fpga = require "mod_fpgaController"
local _done, features = pcall(require, 'features')
if not _done then
	features = {}
elseif type(features) ~= "table" then
	features = _G
end

--  FPGA format register: 0xf2
--	7:6 Video format 				(0-SD, 1-1280x720, 2-1920x1035, 3-1920x1080) 
--	5:3 Frame format 				(0,1,,7 ---> res,24p,25p,30p,50i,60i,50p,60p) 
--	2:fractional rate selection,    (1 : fractional rate(only for 60p、60i、30p、24p) 0 : integral)
--	1:cov hd to sd					(1: cov(only for 1080i50,1080i60) 0: no cov)
--	0:valid or invalid				(1:valid,0:invalid)  由fpga写入1，MCU写入0，
--
--	NOTE: For new version FPGA firmware ( >= D4 ):
--  0xF3:P转I [W] 当需要转换时写入0X01,否则写入0X00
--  0x01: p-i 1080p60 TO 1080i60 1080p50 TO 1080i50
--  0x00:不转换	

local fpgaOutputMapTable = {
    ["PAL"] = { 3, 4, 0, 1 },
    ["576I"] = { 3, 4, 0, 1 },
    ["NTSC"] = { 3, 5, 0, 1 },
    ["480I"] = { 3, 5, 0, 1 },
    ["720P50"] = { 1, 6, 0, 0 },
    ["720P60"] = { 1, 7, 0, 0 },
    ["1080P24"] = { 3, 1, 0, 0 },
    ["1080P25"] = { 3, 2, 0, 0 },
    ["1080P30"] = { 3, 3, 0, 0 },
--  ["1080I50"] = { 3, 4, 0, 0 },
--  ["1080I60"] = { 3, 5, 0, 0 },
--  New method: Input is 1080p but convert to 1080i
    ["1080I50"] = { 3, 6, 0, 0 },
    ["1080I60"] = { 3, 7, 0, 0 },
    ["1080P50"] = { 3, 6, 0, 0 },
    ["1080P60"] = { 3, 7, 0, 0 }
}

local _M = {
}

----------------------------------------------
-- SDI meta.
----------------------------------------------

local sdi_meta = {}

--copy form D300
function sdi_meta:Init()
	if features.FEATURE_DISABLE_SDI_AUDIO_G1P2 then
		fpga.WriteReg( 0xf5, {0x01} )
	else
		fpga.WriteReg( 0xf5, {0x00} )
	end
end

--This is *MUST REQUIRED* for soc/output module.
function sdi_meta:Enable()
end

--This is *MUST REQUIRED* for soc/output module.
function sdi_meta:Disable()
	fpga.WriteReg( 0xf3, {0x00} ) --Close the P->I conversion.
end

--This is *MUST REQUIRED* for soc/output module.
function sdi_meta:UpdateResolution( res, syncDetail, t1001 )
	if not res then
		_dbg.Error( "SDI", "Invalid resolution while updating!" )
		return false
	end

	res = res:upper()
	local regs = fpgaOutputMapTable[res]
	if regs then
		local regVal = regs[1] << 6
		regVal = regVal + (regs[2] << 3)
		if t1001 then
			regVal = regVal + (1<<2)
		end
		regVal = regVal + (regs[4] << 1)
		fpga.WriteReg( 0xf2, {regVal})

		--New method:
		if res == "1080I50" or res == "1080I60" or res == "1080I" then
			fpga.WriteReg( 0xf3, {0x01} )
		else
			fpga.WriteReg( 0xf3, {0x00} )
		end
	else
		_dbg.Error( "SDI", "Invalid/unsupported resolution:", res, "while updating!" )
		return false
	end
end

function sdi_meta:TranslateResolution(res)
	if res then
		if res == "NTSC" or res == "480I" or res == "480I60" then
			return "1080I60"
		elseif res == "PAL" or res == "576I" or res == "576I50" then
			return "1080I50"
		--New method: Convert 1080p->1080i
		elseif res == "1080I50" or res == "1080I" then
			return "1080P50"
		elseif res == "1080I60" then
			return "1080P60"
		else
			return nil
		end
	else
		return nil
	end
end

sdi_meta.__index = sdi_meta

----------------------------------------------

local default_sdi = nil

function _M.Init()
	if not default_sdi then
		default_sdi = {
			interface_type = "SDI"
		}

		setmetatable( default_sdi, sdi_meta )
		default_sdi:Init()
	end
end

--Now only support one instance.
function _M.GetInstance()
	return default_sdi
end

return _M

