local coaf = require "coaf.core"
local _dbg = require "coaf.debug"

local TOOLS = require "system.cmdtools"
local ntpCli = require "system.ntp_client"
local ntpSvr = require "system.ntp_server"
local hwclock = require "system.hwclock"
local timezone = require "system.timezone"

pcall( require, "lsystime" )

local _M = {
	configs = {
		LOCATION = "Asia/Shanghai",
		OFFSET = 8,
		NTP = false,
		NTPD = true,
		NTP_SERVERS = "ntp1.aliyun.com ntp2.aliyun.com ntp3.aliyun.com"
	},
	ntpReady = false
}

-- **configuration**
-- LOCATION="Asia/Shanghai"
-- OFFSET=8
-- NTP=true/false
-- NTP_SERVERS="addr1 addr2 addr3 ..."
local function loadConfig()
	local cfgfile = io.open( "/etc/time.conf", "r" )
	if cfgfile then
		local line = cfgfile:read("*l")
		while line do
			if not line:match("^%s*#") then --Not be comment.
				local key, val = line:match("([%w_]*)%s*=%s*(.+)$" )
				if key then
					local exactVal = val:match("%b''")
					if exactVal then
						val = exactVal:match("'(.*)'")
					else
						exactVal = val:match('%b""')
						if exactVal then
							val = exactVal:match('"(.*)"')
						else
							exactVal = val:match( "(.-)%s*#.*$" )
							if exactVal then
								val = exactVal
							else
								val = val:match("(.-)%s*$")
							end
						end
					end

					if key == "NTP" or key == "NTPD" then
						if val == "true" or val == "yes" then
							_M.configs[key] = true
						else
							_M.configs[key] = false
						end
					elseif key == "OFFSET" then
						_M.configs[key] = tonumber(val) or 8
					else
						_M.configs[key] = val
					end
				end --matched key=value
			end -- not be comment
			line = cfgfile:read("*l")
		end --while
		cfgfile:close()
		return true
	else
		_dbg.Warn( "SYSCTRL", "Fail to load time configuration file: /etc/time.conf" )
		return false
	end
end

local function saveConfig()
	local cfgfile = io.open( "/etc/time.conf", "w" )
	if not cfgfile then
		_dbg.Error( "SYSCTRL", "Fail to save time configuration file: /etc/time.conf" )
		return false
	else
		cfgfile:write( "LOCATION=\"" .. (_M.configs["LOCATION"] or "Asia/Shanghai") .. "\"\n" )
		cfgfile:write( "OFFSET=" .. (_M.configs["OFFSET"] or 8) .. "\n" )
		cfgfile:write( "NTP=" .. (_M.configs["NTP"] and "true" or "false") .. "\n" )
		cfgfile:write( "NTPD=" .. (_M.configs["NTPD"] and "true" or "false") .. "\n" )
		cfgfile:write( "NTP_SERVERS=\"" .. (_M.configs["NTP_SERVERS"] or "") .. "\"\n" )
		cfgfile:close()
		return true
	end
end

function _M.UpdateConfig(initState)
	timezone.Update( _M.configs["LOCATION"] or "NONE", _M.configs["OFFSET"] )
	if _M.configs["NTP"] then
		if initState then --When initializing, load RTC time first.
			hwclock.Load(initState)
		end
		local ntps = _M.configs["NTP_SERVERS"]
		local cmdargs = ""
		ntps = ntps .. " "
		local servers = {}
		local item
		for item in ntps:gmatch("(.-)%s+") do
			if item ~= "" then
				table.insert(servers, item)
			end
		end
		if #servers == 0 then
			ntpCli.Stop()
			hwclock.Load()
			_M.ntpReady = false
		else
			ntpCli.Start(servers)
			_M.ntpReady = true
		end
	else
		ntpCli.Stop()
		hwclock.Load(initState)
		_M.ntpReady = false
	end

	if _M.periodTimer then
		_M.periodTimer:again()
	end

	if _M.configs["NTPD"] then
		ntpSvr.Start()
	else
		ntpSvr.Stop()
	end
end

function _M.SetLocation( loc, offset )
	_M.configs["LOCATION"] = loc
	_M.configs["OFFSET"] = offset or 8
	saveConfig()
	_M.UpdateConfig()
	return true
end

function _M.GetLocation()
	return _M.configs["LOCATION"] or "Asia/Shanghai", _M.configs["OFFSET"] or 8
end

function _M.SetDatetime( dt )
	if dt and dt:match("%d*%-%d*%-%d*%s+%d*:%d*:%d*") then
		if LSYSTIME then
			LSYSTIME.AdjustTime(dt,-1000000,1000000)
		else
			os.execute( TOOLS.SET_DATETIME .. " '" .. dt .. "'" )
		end
		hwclock.Store(true)
		if _M.periodTimer then
			_M.periodTimer:again()
		end
		return true
	else
		return false, "Invalid date/time format"
	end
end

function _M.SyncHwClock(store)
	if store then
		hwclock.Store(true)
	else
		hwclock.Load(true)
		if _M.periodTimer then
			_M.periodTimer:again()
		end
	end
	return true
end

function _M.SetNtp(enable, ntpServers)
	_M.configs["NTP"] = enable
	if ntpServers then
		_M.configs["NTP_SERVERS"] = ntpServers
	end
	saveConfig()
	_M.UpdateConfig()
	return true
end

function _M.GetNtp()
	return _M.configs["NTP"], _M.configs["NTP_SERVERS"]
end

function _M.SetNtpService(enable)
	_M.configs["NTP"] = enable
	saveConfig()
	_M.UpdateConfig()
	return true
end

local function periodTimeSync()
	if not _M.ntpReady then
		--hwclock.Load() --NOTE: Don't load.
	else
		hwclock.Store() --If kernel support ntp->hwclock sync, it's not neccessary too.
	end
end

function _M.Init()
	if LSYSTIME then
		LSYSTIME.SetAllowTimeDiff(100000)	--100ms.
	end

	local r
	loadConfig()
	_M.UpdateConfig(true) --Init
	--_M.periodTimer = coaf.NewTimer( periodTimeSync, 0.01, 5*60 ) -- 5 minutes.
	--if _M.periodTimer then
	--	_M.periodTimer:start()
	--end
end

return _M

