#!/usr/bin/env lua
local coaf = require 'coaf.core'
local _dbg = require 'coaf.debug'

-- _G.TIME_TICK = 0

coaf.ChDir(coaf.GetAppDir(arg))

local socket = require 'cosocket.core'
local client = require 'coaf.client'
local adapter = require 'coaf.cli_adapter.simple'
local service = require 'coaf.service'

adapter.timeoutUnit = 1000
client.InitSocket(nil, adapter)

local withLog = false
local _x
for _x = 1, #arg do
    if arg[_x] == '--log' then
        withLog = true
        break
    end
end

if withLog then
    -- ** initialize the log file **
    local roll = require 'coaf.log.rollfile'
    os.execute('/bin/mkdir -p /var/log/')
    local logF = roll.Create('/var/log/ui.log')
    if not logF or not logF:init() then
        _dbg.Error('UI', 'Fail to open log file: /var/log/ui.log, so the log messages should be output to console.')
        withLog = false
    else
        _dbg.AddLogger('logfile', logF, nil, nil)
    end
-- ** log **
end

coaf.Init('ui')

service.InitSocket()
client.InitSocket(socket, adapter)

local LED = require 'ui_leds'
local SW = require 'ui_switchs'
local CONFIG = require 'ui_config'

_G.config = CONFIG.Load()

-- ** CoService **
local cosvc_ui
local ui_service = require 'ui_service'
------------------

local function GetProxy(objPath, service, loc, timeout, addr)
	local proxy
	local cli

	cli = client.New( service, loc, addr, timeout or 1 ) --3 seconds timeout.
	if not cli then
		_dbg.Error( "UI", "Fail to create connection to service '" .. (loc or "local") .. "/" .. (service or "nil") )
		return nil
	end
	proxy = cli:CreateProxy(objPath, true) --Exclusive
	if not proxy then
		_dbg.Error( "UI", "Fail to create object proxy for '" .. (objPath or "unknown") .. "'" )
		return nil
	end
	return proxy
end

local function GetLinkStatus()
    local network_ok = true
	local proxy = GetProxy('/', 'networkmanager')
    if not proxy then
        network_ok = false
	else
		local r, eth0, eth1 
		r, eth0 = proxy.GetLinkDetail('eth0')
        if not r then
            _dbg.Error('UI', 'Get eth0 LinkDetail error！')
			network_ok = false
		end

		r, eth1 = proxy.GetLinkDetail('eth1')
        if not r then
            _dbg.Error('UI', 'Get eth1 LinkDetail error！')
			network_ok = false
		end

        if eth0.status ~= 'up' and eth1.status ~= 'up' then
            _dbg.Error('UI', 'eth0 and eth1 not up!')
            network_ok = false
		end
		proxy:Destroy()        
    end
    return network_ok
end

local function GetCodecStatus()
	local proxy = GetProxy('/streams', 'codec')
    if not proxy then
        _dbg.Error('UI', 'Can not connect codec service...')
		return false
	else
		local r, streams = proxy.GetStreams()
		proxy:Destroy()
		if not r or not streams then
			return false
        else
            return true
        end
	end
end

local for1s_PeriodCheck_working = false 
local function for1s_PeriodCheck()
	if for1s_PeriodCheck_working then
		return
	end
	for1s_PeriodCheck_working = true
    if GetCodecStatus() then
        LED.SetLightStatus('run', LED.BLINK)
    else
        LED.SetLightStatus('run', LED.OFF)
    end

    if GetLinkStatus() then
        LED.SetLightStatus('link', LED.ON)
    else
        LED.SetLightStatus('link', LED.OFF)
	end
	for1s_PeriodCheck_working = false
end

local function for125ms_PeriodCheck()
    -- _G.TIME_TICK = _G.TIME_TICK + 0.125
    LED.Step()
    SW.Step()
end

-- ** Timers **
local for125msTimer
local for1sTimer

-- **Main entry**
local function Main(args)
    ui_service.Init()
    cosvc_ui = service.New('ui', 'unix:/var/run/ui.sock')
    if cosvc_ui then
        _dbg.Info('UI', "CoService of 'ui' is created.")
        cosvc_ui:CreateCoServerProxy()
        cosvc_ui:ExposeObject('/', ui_service)
        cosvc_ui:ExposeObject('/config', _G.config)
    else
        _dbg.Error('UI', 'Fail created CoService!')
    end

    LED.SetLightStatus('run', LED.OFF)
    LED.SetLightStatus('link', LED.OFF)

    for125msTimer = coaf.NewTimer(for125ms_PeriodCheck, 0.01, 0.125)
    if for125msTimer then
        for125msTimer:start()
    end

    for1sTimer = coaf.NewTimer(for1s_PeriodCheck, 0.1, 5)
    if for1sTimer then
        for1sTimer:start()
    end
end

coaf.Start(Main, arg)

if for125msTimer then
    for125msTimer:stop()
end

if for1sTimer then
    for1sTimer:stop()
end

_dbg.Print('UI', 'Service terminated.')
